/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <sstream>
#include <vector>
#include <cmath>
#include <pugixml.hpp>
#include "Configuration/ConfReader.hpp" //Hikou no mizu's ConfReader, see ../src

/// Struct to contain a <animation> tag, and its associated <hits> tag if any
struct AnimNode
{
    AnimNode()
    {
        
    }
    
    AnimNode(const pugi::xml_node &animElement, const pugi::xml_node &hitsElement) :
    anim(animElement), hits(hitsElement)
    {
        
    }
    
    pugi::xml_node anim, hits;
};

std::vector<AnimNode> parseXML(pugi::xml_document&, const std::string&);
void preprocessAnimNode(pugi::xml_node&);
void updateAnimNode(pugi::xml_node&, int, int, int, int);
void updateHitsNode(pugi::xml_node&, int, int);
const char *roundAsChars(float number);

///descbuild: Converts a single abstract character animation description tag
///into a game-readable format (absolute values for frame positions and sizes)
int main(int argc, char **argv)
{
    if (argc <= 2)
    {
        std::cerr << "Usage: " << argv[0] << " [animation_metadata_file] [dest_xml_description_file]" << std::endl;
        return 1;
    }
    
    std::string animInfoPath(argv[1]);
    const char *xmlDescriptionPath = argv[2];
    
    //Read the metadata of the generated animation.
    ConfReader animInfoReader;
    if (!animInfoReader.load(animInfoPath))
    {
        std::cerr << "Error while loading the file: " << animInfoPath << std::endl;
        return 1;
    }
    
    std::string fullAnimName;
    int framesAmount, frameWidth, frameHeight, tiles;
    
    bool readingSuccess = animInfoReader.getItem("animation_name", fullAnimName)
                       && animInfoReader.getIntItem("frames_amount", framesAmount)
                       && animInfoReader.getIntItem("frame_width", frameWidth)
                       && animInfoReader.getIntItem("frame_height", frameHeight)
                       && animInfoReader.getIntItem("tiles_per_column", tiles);
    if (!readingSuccess)
    {
        std::cerr << "Error while parsing the file: " << animInfoPath << std::endl;
        return 1;
    }
    else if (framesAmount <= 0 || frameWidth <= 0 || frameHeight <= 0 || tiles <= 0)
    {
        std::cerr << "Malformed file, at least one value is not strictly positive: " << animInfoPath << std::endl;
        return 1;
    }
    
    //Extract character and animation name (without the .png extension).
    std::string::size_type delimiterPos = fullAnimName.find('/');
    if (delimiterPos == std::string::npos)
    {
        std::cerr << "Error while parsing the animation name: " << fullAnimName << std::endl;
        return 1;
    }
    
    std::string animName = fullAnimName.substr(delimiterPos+1, fullAnimName.length()-delimiterPos-5);
    
    //Open the xml animations' description file for filling in.
    pugi::xml_document xmlFile;
    if (!xmlFile.load_file(xmlDescriptionPath, pugi::parse_default | pugi::parse_declaration))
    {
        std::cerr << "Error while loading the file: " << xmlDescriptionPath << std::endl;
        return 1;
    }
    
    //Look for the matching <animation> tag.
    std::vector<AnimNode> animNodes = parseXML(xmlFile, animName);
    if (animNodes.empty())
    {
        std::cerr << "Could not find an animation with name: " << animName << std::endl;
        return 1;
    }
    
    //Update the tags.
    std::vector<AnimNode>::iterator it;
    for (it = animNodes.begin(); it != animNodes.end(); it++)
    {
        preprocessAnimNode(it->anim); //Unfold <frameSet> elements
        updateAnimNode(it->anim, framesAmount, frameWidth, frameHeight, tiles);
        if (it->hits)
            updateHitsNode(it->hits, frameWidth, frameHeight);
    }
    
    //Save the updated file and exit.
    xmlFile.save_file(xmlDescriptionPath, "", pugi::format_default, pugi::encoding_utf8);
    
    return 0;
}

///Parses the input 'xmlFile' looking for the description of the requested animation
std::vector<AnimNode> parseXML(pugi::xml_document &xmlFile, const std::string &animName)
{
    std::vector<AnimNode> animNodes;
    
    pugi::xml_node stateNode;
    for (stateNode = xmlFile.child("main").child("data").first_child();
         stateNode;
         stateNode = stateNode.next_sibling())
    {
        //Main animation?
        pugi::xml_node animNode = stateNode.child("animation");
        if (std::string(animNode.attribute("path").value()) == animName)
            animNodes.push_back( AnimNode(animNode, pugi::xml_node()) ); //A main animation: No associated <hits> tag.
        else
        {
            //Attacks?
            pugi::xml_node attackNode;
            for (attackNode = stateNode.child("attacks").first_child();
                 attackNode;
                 attackNode = attackNode.next_sibling())
            {
                pugi::xml_node attackAnimNode = attackNode.child("animation");
                if (std::string(attackAnimNode.attribute("path").value()) == animName)
                    animNodes.push_back( AnimNode(attackAnimNode, attackNode.child("hits")) );
            }
        }
    }
    
    return animNodes;
}

///Preprocesses a <animation> tag to unfold the <frameSet> children into sets of <frame> children
void preprocessAnimNode(pugi::xml_node &animNode)
{
    pugi::xml_node frameSetNode = animNode.child("frameSet");
    while (frameSetNode)
    {
        //Extract the frame range and body coordinates of the <frameSet> element
        std::string rangeStr(frameSetNode.attribute("id_range").value());
        
        const char *relBodyX = frameSetNode.attribute("bodyX_rel").value(),
                   *relBodyY = frameSetNode.attribute("bodyY_rel").value(),
                   *relBodyWidth = frameSetNode.attribute("bodyWidth_rel").value(),
                   *relBodyHeight = frameSetNode.attribute("bodyHeight_rel").value();
        
        std::string::size_type delimiterPos = rangeStr.find(':');
        if (delimiterPos != std::string::npos)
        {
            int startFrame, endFrame;
            std::istringstream(rangeStr.substr(0, delimiterPos)) >> startFrame;
            std::istringstream(rangeStr.substr(delimiterPos+1, rangeStr.length()-delimiterPos-1)) >> endFrame;
            
            //Unfold the <frameSet> element
            for (int frameId = ((startFrame <= endFrame) ? startFrame : endFrame);
                frameId <= ((startFrame <= endFrame) ? endFrame : startFrame);
                frameId++)
            {
                //Insert the <frame> element before or after the <frameSet> element
                //depending on the wanted order
                pugi::xml_node frame;
                if (startFrame <= endFrame) frame = animNode.insert_child_before("frame", frameSetNode);
                else frame = animNode.insert_child_after("frame", frameSetNode);
                
                frame.append_attribute("id").set_value(frameId);
                frame.append_attribute("bodyX_rel").set_value(relBodyX);
                frame.append_attribute("bodyY_rel").set_value(relBodyY);
                frame.append_attribute("bodyWidth_rel").set_value(relBodyWidth);
                frame.append_attribute("bodyHeight_rel").set_value(relBodyHeight);
            }
        }
        
        animNode.remove_child(frameSetNode);
        frameSetNode = animNode.child("frameSet");
    }
}

///Updates a <animation> tag from relative values to absolute, game-readable values
void updateAnimNode(pugi::xml_node &animNode, int framesAmount, int frameWidth, int frameHeight, int tiles)
{
    //The number of horizontal/vertical pixels between each frame.
    constexpr int framesGap = 4;
    
    //Compute the position of each frame in the spritesheet.
    std::vector< std::pair<int, int> > frames;
    for (int i = 0; i < framesAmount; i++)
    {
        int colIx = i%tiles, rowIx = static_cast<int>(i/tiles); // tiles > 0
        int frameX = colIx * (frameWidth + framesGap), frameY = rowIx * (frameHeight + framesGap);
        
        frames.push_back(std::make_pair(frameX, frameY));
    }
    
    pugi::xml_node frameNode;
    for (frameNode = animNode.child("frame");
         frameNode;
         frameNode = frameNode.next_sibling("frame"))
    {
        //Query abstract frame description attributes
        int frameId = frameNode.attribute("id").as_int(-1);
        
        float relBodyX = frameNode.attribute("bodyX_rel").as_float(0.f),
              relBodyY = frameNode.attribute("bodyY_rel").as_float(0.f),
              relBodyWidth = frameNode.attribute("bodyWidth_rel").as_float(1.f),
              relBodyHeight = frameNode.attribute("bodyHeight_rel").as_float(1.f);
        
        //Add static frame data
        if (frameId >= 0 && frameId < framesAmount)
        {
            //Frame absolute position and size
            unsigned int uFrameId = static_cast<unsigned int>(frameId);
            frameNode.append_attribute("x").set_value(frames[uFrameId].first);
            frameNode.append_attribute("y").set_value(frames[uFrameId].second);
            frameNode.append_attribute("width").set_value(frameWidth);
            frameNode.append_attribute("height").set_value(frameHeight);
            
            //Absolute values for body x, y, width and height
            frameNode.append_attribute("bodyX").set_value( roundAsChars(relBodyX * static_cast<float>(frameWidth)) );
            frameNode.append_attribute("bodyY").set_value( roundAsChars(relBodyY * static_cast<float>(frameHeight)) );
            frameNode.append_attribute("bodyWidth").set_value( roundAsChars(relBodyWidth * static_cast<float>(frameWidth)) );
            frameNode.append_attribute("bodyHeight").set_value( roundAsChars(relBodyHeight * static_cast<float>(frameHeight)) );
            
            //Remove abstract frame description attributes
            frameNode.remove_attribute("id");
            frameNode.remove_attribute("bodyX_rel");
            frameNode.remove_attribute("bodyY_rel");
            frameNode.remove_attribute("bodyWidth_rel");
            frameNode.remove_attribute("bodyHeight_rel");
        }
    }
}

///Updates a <hits> tag from relative values to absolute, game-readable values
void updateHitsNode(pugi::xml_node &hitsNode, int frameWidth, int frameHeight)
{
    pugi::xml_node hitNode;
    for (hitNode = hitsNode.child("hit");
         hitNode;
         hitNode = hitNode.next_sibling("hit"))
    {
        float relX = hitNode.attribute("x_rel").as_float(-1.f),
              relY = hitNode.attribute("y_rel").as_float(-1.f),
              relWidth = hitNode.attribute("width_rel").as_float(-1.f),
              relHeight = hitNode.attribute("height_rel").as_float(-1.f);
        
        if (relX > 0.f && relY > 0.f)
        {
            hitNode.append_attribute("x").set_value( roundAsChars(relX * static_cast<float>(frameWidth)) );
            hitNode.append_attribute("y").set_value( roundAsChars(relY * static_cast<float>(frameHeight)) );
            hitNode.remove_attribute("x_rel");
            hitNode.remove_attribute("y_rel");
            
            if (relWidth > 0.f && relHeight > 0.f)
            {
                hitNode.append_attribute("width").set_value( roundAsChars(relWidth * static_cast<float>(frameWidth)) );
                hitNode.append_attribute("height").set_value( roundAsChars(relHeight * static_cast<float>(frameHeight)) );
                hitNode.remove_attribute("width_rel");
                hitNode.remove_attribute("height_rel");
            }
        }
    }
}

///Rounds 'number' to 0.1 precision and returns the result as a string
const char *roundAsChars(float number)
{
    std::ostringstream ss;
    ss << roundf(number * 10.f) / 10.f;
    
    return ss.str().c_str();
}
