/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Animation.hpp"

Animation::Animation() : Drawable(), m_texture(nullptr),
m_frameIx(0), m_duration(0.f), m_time(0.f),
m_x1(0.f), m_y1(0.f), m_x2(0.f), m_y2(0.f)
{
    
}

void Animation::draw()
{
    //Update matrix
    Drawable::pushMatrix();
    Drawable::updateMatrix();
    
    if (m_texture != nullptr && m_frameIx < m_framesList.size())
        m_texture->draw(m_x1, m_y1, m_x2, m_y2);
    
    Drawable::popMatrix();
}

void Animation::addFrame(const Frame &frame)
{
    m_framesList.push_back(frame);
    computeTextureCoordinates();
}

void Animation::goToTime(float time)
{
    if (m_duration > 0.f && (time >= 0.f && time <= m_duration))
    {
        m_time = time;
        
        unsigned int frameIx = static_cast<unsigned int>(
            static_cast<float>(m_framesList.size()) * (m_time / m_duration)
        );
        
        //Recompute texture coordinates if the frame changed
        if (m_frameIx != frameIx && frameIx < m_framesList.size())
        {
            m_frameIx = frameIx;
            computeTextureCoordinates();
        }
    }
}

void Animation::goForward(float distance)
{
    float bufferTime = m_time + distance;
    
    if (bufferTime > m_duration)
        bufferTime = m_duration;
    
    goToTime(bufferTime);
}

void Animation::goBack(float distance)
{
    float bufferTime = m_time - distance;
    
    if (bufferTime < 0.f)
        bufferTime = 0.f;
    
    goToTime(bufferTime);
}

void Animation::computeTextureCoordinates()
{
    if (m_texture != nullptr && m_frameIx < m_framesList.size())
    {
        GLfloat textureWidth = static_cast<GLfloat>(m_texture->getWidth()),
                textureHeight = static_cast<GLfloat>(m_texture->getHeight());
        
        const Box &frame = m_framesList[m_frameIx].srcBox;
        
        m_x1 = (textureWidth > 0.f) ? static_cast<GLfloat>(frame.left) / textureWidth : 0.f;
        m_y1 = (textureHeight > 0.f) ? static_cast<GLfloat>(frame.top) / textureHeight : 0.f;
        
        m_x2 = (textureWidth > 0.f) ? static_cast<GLfloat>(frame.left + frame.width) / textureWidth : 1.f;
        m_y2 = (textureHeight > 0.f) ? static_cast<GLfloat>(frame.top + frame.height) / textureHeight : 1.f;
    }
}

//Get and set methods
unsigned int Animation::getFrameSize() const
{
    return static_cast<unsigned int>(m_framesList.size());
}

void Animation::setTexture(Texture &texture)
{
    m_texture = &texture;
    computeTextureCoordinates();
}

const Texture &Animation::getTexture() const
{
    return *m_texture;
}

void Animation::setDuration(float duration)
{
    if (duration > 0.f)
        m_duration = duration;
}

float Animation::getDuration() const
{
    return m_duration;
}

float Animation::getCurrentTime() const
{
    return m_time;
}

bool Animation::atInitialState() const
{
    return (m_time <= FLOAT_ERROR_ANIM);
}

bool Animation::atEndedState() const
{
    return (m_time >= m_duration - FLOAT_ERROR_ANIM);
}

bool Animation::getBodyBox(Box &target) const
{
    if (m_texture != nullptr && m_frameIx < m_framesList.size())
    {
        target = m_framesList[m_frameIx].bodyBox;
        return true;
    }
    
    return false;
}

float Animation::getWidth() const
{
    if (m_texture != nullptr && m_frameIx < m_framesList.size())
        return m_framesList[m_frameIx].srcBox.width;
    
    return 0.f;
}

float Animation::getHeight() const
{
    if (m_texture != nullptr && m_frameIx < m_framesList.size())
        return m_framesList[m_frameIx].srcBox.height;
    
    return 0.f;
}
