/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Arena.hpp"

Arena::Arena() : m_width(0.f), m_height(0.f), m_ground(0.f), m_groundMaterial(MATERIAL_NONE),
m_gravity(0.f), m_airFriction(0.f), m_minCameraWidth(0.f)
{
    
}

Arena::Arena(const Sprite &background, float width, float height, float ground, float gravity, float airFriction) :
m_background(background), m_width(width), m_height(height), m_ground(ground), m_groundMaterial(MATERIAL_NONE),
m_gravity(gravity), m_airFriction(airFriction), m_minCameraWidth(1920.f)
{
    resizeBackground();
}

Arena::~Arena()
{
    removePlatforms();
}

void Arena::draw()
{
    m_background.draw();
    
    std::vector< std::pair<Platform*, Box> >::iterator it;
    
    for (it = m_platformsList.begin(); it != m_platformsList.end(); it++)
        it->first->draw();
}

void Arena::addPlatform(const Box &spawnBox, const Box &bodyBox, const Sprite &sprite,
                        const std::string &material, bool hitable, bool fixed)
{
    Platform *platform = new Platform();
    platform->setBodyBox(bodyBox);
    platform->setSprite(sprite);
    platform->setGraphicsSize(spawnBox.width, spawnBox.height);
    platform->setMaterial(material);
    platform->setHitable(hitable);
    platform->setFixed(fixed);
    
    m_platformsList.push_back( std::make_pair(platform, spawnBox) );
}

void Arena::removePlatform(unsigned int platformIx)
{
    if (platformIx < m_platformsList.size())
    {
        delete m_platformsList[platformIx].first;
        m_platformsList.erase(m_platformsList.begin() + platformIx);
    }
}

void Arena::removePlatforms()
{
    std::vector< std::pair<Platform*, Box> >::iterator it;
    
    for (it = m_platformsList.begin(); it != m_platformsList.end(); it++)
        delete it->first;
    
    m_platformsList.clear();
}

void Arena::applyToPhysics(PhysicsWorld &physicsWorld)
{
    //Configure a physics world to match with the arena parameters
    physicsWorld.setWorldBox( Box(0.f, 0.f, m_width, m_ground) );
    physicsWorld.setGravityForce(m_gravity);
    physicsWorld.setAirFriction(m_airFriction);
    physicsWorld.setGroundMaterial(m_groundMaterial);
}

void Arena::subscribeObjects(PhysicsWorld &physicsWorld, SoundEngine *soundEngine)
{
    //Subscribe arena's platforms to a physics world
    std::vector< std::pair<Platform*, Box> >::iterator it;
    
    for (it = m_platformsList.begin(); it != m_platformsList.end(); it++)
    {
        Platform *platform = it->first;
        const Box &spawnBox = it->second;
        const Box &bodyBox = it->first->getBodyBox();
        
        platform->setPhysicsWorld(physicsWorld);
        platform->setWidth(bodyBox.width);
        platform->setHeight(bodyBox.height);
        platform->setPosition(spawnBox.left, spawnBox.top - platform->getBox().height);
        
        if (soundEngine != nullptr)
            platform->setSoundEngine((*soundEngine));
        
        physicsWorld.addObject((*platform));
    }
}

void Arena::loadFromXML(const std::string &name, const std::string &xmlRelPath, TextureManager &textureManager)
{
    //Delete previously loaded data.
    removePlatforms();

    //TinyXML initialization
    std::string xmlPath = BuildValues::data(xmlRelPath); //Locate absolute path to xml data
    
    pugi::xml_document xmlFile;
    xmlFile.load_file(xmlPath.c_str());
    
    pugi::xml_node arenaNode;
    for (arenaNode = xmlFile.child("main").child("arena");
         arenaNode;
         arenaNode = arenaNode.next_sibling("arena"))
    {
        std::string itemName(arenaNode.attribute("pathname").value());
        if (itemName == name)
        {
            //Get main information
            Texture *texture = textureManager.getTexture("gfx/arenas/" + itemName + ".png");
            float textureWidth = static_cast<float>(texture->getWidth()),
                  textureHeight = static_cast<float>(texture->getHeight());
            
            float width = arenaNode.attribute("width").as_float(1920.f),
                  height = arenaNode.attribute("height").as_float(1080.f),
                  ground = arenaNode.attribute("ground").as_float(1000.f),
                  gravity = arenaNode.attribute("gravity").as_float(2000.f),
                  airFriction = arenaNode.attribute("airFriction").as_float(2250.f),
                  minCameraWidth = arenaNode.attribute("minCameraWidth").as_float(1920.f);
            
            std::string groundMaterial(arenaNode.attribute("material").as_string(MATERIAL_NONE));
            
                //Update arena
            setWidth(width);
            setHeight(height);
            setGround(ground);
            setGroundMaterial(groundMaterial);
            setGravity(gravity);
            setAirFriction(airFriction);
            setMinCameraWidth(minCameraWidth);
            
            //Get detailed information
                //Background
            pugi::xml_node backgroundNode = arenaNode.child("background");
            if (backgroundNode)
            {
                //Get information
                int srcX = backgroundNode.attribute("srcX").as_int(0),
                    srcY = backgroundNode.attribute("srcY").as_int(0),
                    srcWidth = backgroundNode.attribute("srcWidth").as_int(0),
                    srcHeight = backgroundNode.attribute("srcHeight").as_int(0);
                
                //Create and add background to the arena
                Box subRect;
                
                subRect.left = (textureWidth > 0.f) ? static_cast<float>(srcX) / textureWidth : 0.f;
                subRect.top = (textureHeight > 0.f) ? static_cast<float>(srcY) / textureHeight : 0.f;
                
                subRect.width = (textureWidth > 0.f) ? static_cast<float>(srcWidth) / textureWidth : 1.f;
                subRect.height = (textureHeight > 0.f) ? static_cast<float>(srcHeight) / textureHeight : 1.f;
                
                Sprite background;
                background.setTexture(*texture);
                background.setSubRect(subRect);
                
                setBackground(background);
            }
            
                //Platforms
            pugi::xml_node platformNode;
            for (platformNode = arenaNode.child("platforms").child("platform");
                 platformNode;
                 platformNode = platformNode.next_sibling("platform"))
            {
                //Get information
                    //Coords
                float x = platformNode.attribute("x").as_float(0.f),
                      y = platformNode.attribute("y").as_float(0.f),
                      width = platformNode.attribute("width").as_float(0.f),
                      height = platformNode.attribute("height").as_float(0.f);
                
                    //Sprite
                int srcX = platformNode.attribute("srcX").as_int(0),
                    srcY = platformNode.attribute("srcY").as_int(0),
                    srcWidth = platformNode.attribute("srcWidth").as_int(0),
                    srcHeight = platformNode.attribute("srcHeight").as_int(0);
                
                float bodyX = platformNode.attribute("bodyX").as_float(0.f),
                      bodyY = platformNode.attribute("bodyY").as_float(0.f),
                      bodyWidth = platformNode.attribute("bodyWidth").as_float(0.f),
                      bodyHeight = platformNode.attribute("bodyHeight").as_float(0.f);
                
                    //Material?
                std::string material(platformNode.attribute("material").as_string(MATERIAL_NONE));
                
                    //Hitable? Fixed?
                bool hitable = platformNode.attribute("hitable").as_bool(false),
                     fixed = platformNode.attribute("fixed").as_bool(false);
                
                Box subRect;
                
                subRect.left = (textureWidth > 0.f) ? static_cast<float>(srcX) / textureWidth : 0.f;
                subRect.top = (textureHeight > 0.f) ? static_cast<float>(srcY) / textureHeight : 0.f;
                
                subRect.width = (textureWidth > 0.f) ? static_cast<float>(srcWidth) / textureWidth : 1.f;
                subRect.height = (textureHeight > 0.f) ? static_cast<float>(srcHeight) / textureHeight : 1.f;
                
                Sprite sprite;
                sprite.setTexture(*texture);
                sprite.setSubRect(subRect);
                
                //Create and add platform to the arena
                addPlatform( Box(x, y, width, height), Box(bodyX, bodyY, bodyWidth, bodyHeight),
                             sprite, material, hitable, fixed );
            }
            
                //Spawns
            pugi::xml_node spawnNode;
            for (spawnNode = arenaNode.child("spawns").child("spawn");
                 spawnNode;
                 spawnNode = spawnNode.next_sibling("spawn"))
            {
                //Get spawn information
                int playerIx = spawnNode.attribute("playerIx").as_int(0);
                float x = spawnNode.attribute("x").as_float(0.f),
                      y = spawnNode.attribute("y").as_float(0.f);
                
                //Add it to the arena
                setPlayerSpawn(static_cast<unsigned int>(playerIx), Vector(x, y));
            }
        }
    }
}

//Get and set methods
    //Graphics
void Arena::setBackground(const Sprite &background)
{
    m_background = background;
    resizeBackground();
}

const Sprite &Arena::getBackground() const
{
    return m_background;
}

    //Physics
void Arena::setWidth(float width)
{
    m_width = width;
    resizeBackground();
}

float Arena::getWidth() const
{
    return m_width;
}

void Arena::setHeight(float height)
{
    m_height = height;
    resizeBackground();
}

float Arena::getHeight() const
{
    return m_height;
}

void Arena::setGround(float ground)
{
    m_ground = ground;
}

float Arena::getGround() const
{
    return m_ground;
}

void Arena::setGroundMaterial(const std::string &material)
{
    m_groundMaterial = material;
}

const std::string &Arena::getGroundMaterial() const
{
    return m_groundMaterial;
}

void Arena::setGravity(float gravity)
{
    m_gravity = gravity;
}

float Arena::getGravity() const
{
    return m_gravity;
}

void Arena::setAirFriction(float airFriction)
{
    m_airFriction = airFriction;
}

float Arena::getAirFriction() const
{
    return m_airFriction;
}

void Arena::setMinCameraWidth(float minWidth)
{
    if (minWidth > 0.f)
        m_minCameraWidth = minWidth;
}

float Arena::getMinCameraWidth() const
{
    return m_minCameraWidth;
}

    //Others
void Arena::setPlayerSpawn(unsigned int ix, const Vector &playerSpawn)
{
    m_playerSpawns[ix] = playerSpawn;
}

bool Arena::getPlayerSpawn(unsigned int ix, Vector &target) const
{
    std::map<unsigned int, Vector>::const_iterator it = m_playerSpawns.find(ix);
    if (it != m_playerSpawns.end())
    {
        target = it->second;
        return true;
    }
    
    return false;
}

void Arena::resizeBackground()
{
    //Resize background to arena size
    if (m_background.getWidth() > 0.f && m_background.getHeight() > 0.f)
    {
        m_background.setXScale(m_width / m_background.getWidth());
        m_background.setYScale(m_height / m_background.getHeight());
    }
}
