/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysicsObject.hpp"

PhysicsObject::PhysicsObject() :
m_physicsWorld(nullptr), m_airPenetration(0.f), m_material(MATERIAL_NONE), m_fixed(false)
{
    
}

PhysicsObject::~PhysicsObject()
{
    
}

void PhysicsObject::setPhysicsWorld(PhysicsWorld &physicsWorld)
{
    m_physicsWorld = &physicsWorld;
}

PhysicsWorld &PhysicsObject::getPhysicsWorld() const
{
    return (*m_physicsWorld);
}

void PhysicsObject::setPosition(float xPosition, float yPosition)
{
    if (m_physicsWorld != nullptr)
    {
        Box initialBox = m_box;
        
        m_box.left = xPosition;
        m_box.top = yPosition;
        
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (!worldBox.contains(m_box) || hitTest().size() > 0)
            m_box = initialBox;
    }
}

void PhysicsObject::moveXPosition(float distance, bool push)
{
    if (m_physicsWorld != nullptr && (distance > FLOAT_ERROR_PHYS || distance < -FLOAT_ERROR_PHYS))
    {
        // To avoid crossing others object without noticing,
        // We make the object move by steps of its width,
        // So it covers the whole area it's moving through.
        
        int stepsNo_signed = static_cast<int>(distance / m_box.width); /// Number of times it will move by its width.
        float remainder = distance - static_cast<float>(stepsNo_signed) * m_box.width; /// Remainder to move to reach the expected distance.
        
        int stepsNo = (stepsNo_signed > 0) ? stepsNo_signed : -stepsNo_signed;
        for (int i = 0; i < stepsNo; i++)
        {
            if (!attempt_moveXPosition((stepsNo_signed > 0) ? m_box.width : -m_box.width, push))
                return;
        }
        
        attempt_moveXPosition(remainder, push);
    }
}

bool PhysicsObject::attempt_moveXPosition(float distance, bool push)
{
    if (m_physicsWorld != nullptr && (distance > FLOAT_ERROR_PHYS || distance < -FLOAT_ERROR_PHYS))
    {
        float initialPosition = m_box.left;
        
        std::vector<PhysicsObject*> initialHits = hitTest();
        m_box.left += distance;
        
        //Objects
        std::vector<PhysicsObject*> finalHits = hitTest();
        PhysicsRoutines::soustractVectors(finalHits, initialHits);
        
        if (push)
        {
            std::vector<PhysicsObject*>::iterator it;
            
            for (it = finalHits.begin(); it != finalHits.end(); it++) //Move hit objects
            {
                PhysicsObject *hitObject = (*it);
                
                if (!hitObject->isFixed())
                {
                    const Box &hitBox = hitObject->getBox();
                    
                    if (distance > 0.f)
                        hitObject->moveXPosition((m_box.left + m_box.width) - hitBox.left + FLOAT_ERROR_PHYS, true);
                    else if (distance < 0.f)
                        hitObject->moveXPosition(m_box.left - (hitBox.left + hitBox.width) - FLOAT_ERROR_PHYS, true);
                }
            }
            
            //Update hits
            finalHits = hitTest();
            PhysicsRoutines::soustractVectors(finalHits, initialHits);
        }
        
        if (finalHits.size() > 0)
        {
            m_box.left = initialPosition; //Return to initial position.
            
            int checkedSide = 0;
            if (distance < 0.f) checkedSide = 1;
            
            PhysicsRoutines::sortObjectsByX(finalHits, checkedSide);
            
            // 'Collide' callback
            for (std::vector<PhysicsObject*>::iterator it = finalHits.begin(); it != finalHits.end(); it++)
            {
                (*it)->collide( (*this) ); //The object collided with myself
                collide( (*(*it)) ); //I collided with the object
            }
            
            // Attempt to fix the position.
            PhysicsObject *nearestObject = finalHits.front();
            if (distance < 0.f) nearestObject = finalHits.back();
            
            const Box &hitBox = nearestObject->getBox();
            
            if (distance > 0.f)
                setPosition(hitBox.left - m_box.width - FLOAT_ERROR_PHYS, m_box.top);
            else if (distance < 0.f)
                setPosition(hitBox.left + hitBox.width + FLOAT_ERROR_PHYS, m_box.top);
            
            return false; /// Move was corrected not to induce a collision.
        }
        
        //World
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (!worldBox.contains(m_box))
        {
            m_box.left = initialPosition; //Return to initial position.
            
            // Attempt to fix the position.
            if (distance > 0.f)
                setPosition(worldBox.left + worldBox.width - m_box.width, m_box.top);
            else if (distance < 0.f)
                setPosition(worldBox.left, m_box.top);
            
            collideWorld(false); //Notify collision with the world box (not the ground).
            
            return false; /// Move was corrected not to induce a collision.
        }
        
        return true; /// Move happened as asked.
    }
    
    return false;
}

void PhysicsObject::moveYPosition(float distance, bool push)
{
    if (m_physicsWorld != nullptr && (distance > FLOAT_ERROR_PHYS || distance < -FLOAT_ERROR_PHYS))
    {
        // To avoid crossing others object without noticing,
        // We make the object move by steps of its width,
        // So it covers the whole area it's moving through.
        int stepsNo_signed = static_cast<int>(distance / m_box.height); /// Number of times it will move by its height.
        float remainder = distance - static_cast<float>(stepsNo_signed) * m_box.height; /// Remainder to move to reach the expected distance.
        
        int stepsNo = (stepsNo_signed > 0) ? stepsNo_signed : -stepsNo_signed;
        for (int i = 0; i < stepsNo; i++)
        {
            if (!attempt_moveYPosition((stepsNo_signed > 0) ? m_box.height : -m_box.height, push))
                return;
        }
        
        attempt_moveYPosition(remainder, push);
    }
}

bool PhysicsObject::attempt_moveYPosition(float distance, bool push)
{
    if (m_physicsWorld != nullptr && (distance > FLOAT_ERROR_PHYS || distance < -FLOAT_ERROR_PHYS))
    {
        float initialPosition = m_box.top;
        
        std::vector<PhysicsObject*> initialHits = hitTest();
        m_box.top += distance;
        
        //Objects
        std::vector<PhysicsObject*> finalHits = hitTest();
        PhysicsRoutines::soustractVectors(finalHits, initialHits);
        
        if (push)
        {
            std::vector<PhysicsObject*>::iterator it;
            
            for (it = finalHits.begin(); it != finalHits.end(); it++) //Move hit objects
            {
                PhysicsObject *hitObject = (*it);
                
                if (!hitObject->isFixed())
                {
                    const Box &hitBox = hitObject->getBox();
                    
                    if (distance > 0.f)
                        hitObject->moveYPosition((m_box.top + m_box.height) - hitBox.top + FLOAT_ERROR_PHYS, true);
                    else if (distance < 0.f)
                        hitObject->moveYPosition(m_box.top - (hitBox.top + hitBox.height) - FLOAT_ERROR_PHYS, true);
                }
            }
            
            //Update hits
            finalHits = hitTest();
            PhysicsRoutines::soustractVectors(finalHits, initialHits);
        }
        
        if (finalHits.size() > 0)
        {
            m_box.top = initialPosition; //Return to initial position.
            
            int checkedSide = 0;
            if (distance < 0.f) checkedSide = 1;
            
            PhysicsRoutines::sortObjectsByY(finalHits, checkedSide);
            
            // 'Collide' callback
            for (std::vector<PhysicsObject*>::iterator it = finalHits.begin(); it != finalHits.end(); it++)
            {
                (*it)->collide( (*this) ); //The object collided with myself
                collide( (*(*it)) ); //I collided with the object
            }
            
            // Attempt to fix the position.
            PhysicsObject *nearestObject = finalHits.front();
            if (distance < 0.f) nearestObject = finalHits.back();
            
            const Box &hitBox = nearestObject->getBox();
            
            if (distance > 0.f)
                setPosition(m_box.left, hitBox.top - m_box.height - FLOAT_ERROR_PHYS);
            else if (distance < 0.f)
                setPosition(m_box.left, hitBox.top + hitBox.height + FLOAT_ERROR_PHYS);
            
            return false; /// Move was corrected not to induce a collision.
        }
        
        //World
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (!worldBox.contains(m_box))
        {
            m_box.top = initialPosition; //Return to initial position.
            
            if (distance > 0.f)
            {
                setPosition(m_box.left, worldBox.top + worldBox.height - m_box.height);
                collideWorld(true); //Notify collision with the world box (the ground).
            }
            else if (distance < 0.f)
            {
                setPosition(m_box.left, worldBox.top);
                collideWorld(false); //Notify collision with the world box (the ceiling).
            }
            
            return false; /// Move was corrected not to induce a collision.
        }
        
        return true; /// Move happened as asked.
    }
    
    return false;
}

void PhysicsObject::setWidth(float width, bool reverseSide)
{
    if (m_physicsWorld != nullptr && width > 0.f)
    {
        Box initialBox = m_box;
        m_box.width = width;
        if (reverseSide) m_box.left += initialBox.width - m_box.width;
        
        //Lower width : no collision, change & abort
        if (initialBox.width >= width)
            return;
        
        //Leave world : abort
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (!worldBox.contains(m_box))
        {
            m_box = initialBox;
            return;
        }
        
        //Potential Object hit
        std::vector<PhysicsObject*> initialHits = hitTest(initialBox);
        std::vector<PhysicsObject*> finalHits = hitTest();
        
        PhysicsRoutines::soustractVectors(finalHits, initialHits);
        
        if (finalHits.size() > 0) //Hits
        {
            // If fixed objects are hit abort to avoid moving objects past the fixed objects.
            if (PhysicsRoutines::containsFixedObject(finalHits))
            {
                m_box = initialBox;
                return;
            }
            
            // Otherwise, try to move the objects.
            std::vector<PhysicsObject*>::iterator it;
            
            for (it = finalHits.begin(); it != finalHits.end(); it++) //Move hit objects
            {
                if (!(*it)->isFixed())
                {
                    const Box &hitBox = (*it)->getBox();
                    
                    if (!reverseSide)
                        (*it)->moveXPosition((m_box.left + m_box.width) - hitBox.left + FLOAT_ERROR_PHYS, true);
                    else
                        (*it)->moveXPosition(m_box.left - (hitBox.left + hitBox.width) - FLOAT_ERROR_PHYS, true);
                }
            }
            
            // If still hit (one of the objects hit a fixed object for example) : abort
            finalHits = hitTest();
            PhysicsRoutines::soustractVectors(finalHits, initialHits);
            
            if (finalHits.size() > 0)
                m_box = initialBox;
        }
    }
}

void PhysicsObject::setHeight(float height, bool reverseSide)
{
    if (m_physicsWorld != nullptr && height > 0.f)
    {
        Box initialBox = m_box;
        
        m_box.height = height;
        if (reverseSide) m_box.top += initialBox.height - m_box.height;
        
        //Lower height : no collision, abort
        if (initialBox.height >= height)
            return;
        
        //Leave world : abort
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (!worldBox.contains(m_box))
        {
            m_box = initialBox;
            return;
        }
        
        //Potential Object hit
        std::vector<PhysicsObject*> initialHits = hitTest(initialBox);
        std::vector<PhysicsObject*> finalHits = hitTest();
        
        PhysicsRoutines::soustractVectors(finalHits, initialHits);
        
        if (finalHits.size() > 0) //Hits
        {
            // If fixed objects are hit abort to avoid moving objects past the fixed objects.
            if (PhysicsRoutines::containsFixedObject(finalHits))
            {
                m_box = initialBox;
                return;
            }
            
            // Otherwise, try to move the objects.
            std::vector<PhysicsObject*>::iterator it;
            
            for (it = finalHits.begin(); it != finalHits.end(); it++) //Move hit objects
            {
                if (!(*it)->isFixed())
                {
                    const Box &hitBox = (*it)->getBox();
                    
                    if (!reverseSide)
                        (*it)->moveYPosition((m_box.top + m_box.height) - hitBox.top + FLOAT_ERROR_PHYS, true);
                    else
                        (*it)->moveYPosition(m_box.top - (hitBox.top + hitBox.height) - FLOAT_ERROR_PHYS, true);
                }
            }
            
            // If still hit (one of the objects hit a fixed object for example) : abort
            finalHits = hitTest();
            PhysicsRoutines::soustractVectors(finalHits, initialHits);
            
            if (finalHits.size() > 0)
                m_box = initialBox;
        }
    }
}

void PhysicsObject::setXVelocity(float xVelocity)
{
    m_velocity.x = xVelocity;
}

void PhysicsObject::setYVelocity(float yVelocity)
{
    m_velocity.y = yVelocity;
}

void PhysicsObject::setVelocity(const Vector &velocity)
{
    if (!isFixed())
        m_velocity = velocity;
}

const Vector &PhysicsObject::getVelocity() const
{
    return m_velocity;
}

void PhysicsObject::setAirPenetration(float airPenetration)
{
    if (airPenetration >= 0.f && airPenetration <= 1.f)
        m_airPenetration = airPenetration;
}

float PhysicsObject::getAirPenetration() const
{
    return m_airPenetration;
}

void PhysicsObject::setMaterial(const std::string &material)
{
    m_material = material;
}

const std::string &PhysicsObject::getMaterial() const
{
    return m_material;
}

void PhysicsObject::setFixed(bool fixed)
{
    m_fixed = fixed;
}

bool PhysicsObject::isFixed() const
{
    return m_fixed;
}

float PhysicsObject::shadowProjection() const
{
    if (m_physicsWorld == nullptr)
        return 0.f;
    
    // The object touches the ground
    const Box &worldBox = m_physicsWorld->getWorldBox();
    if (m_box.top + m_box.height >= worldBox.top + worldBox.height)
        return 0.f;
    
    std::vector<PhysicsObject*> objectsUnder = hitTest(
        Box(m_box.left + m_box.width*.2f, m_box.top, m_box.width*.6f,
            worldBox.top + worldBox.height - m_box.top)
        );
    PhysicsRoutines::sortObjectsByY(objectsUnder, 0);
    
    // The object is over another object
    std::vector<PhysicsObject*>::iterator it;
    for (it = objectsUnder.begin(); it != objectsUnder.end(); it++)
    {
        if ((*it)->interceptsShadows())
            return (*it)->getBox().top - (m_box.top + m_box.height);
    }
    
    //The shadow should be projected on the ground
    return worldBox.top + worldBox.height - (m_box.top + m_box.height);
}

/// Whether the object should intercept shadows of other objects over it
bool PhysicsObject::interceptsShadows() const
{
    return false;
}

bool PhysicsObject::isOnGround(std::string *material) const
{
    if (m_physicsWorld != nullptr)
    {
        const Box &worldBox = m_physicsWorld->getWorldBox();
        
        if (m_box.top + m_box.height + FLOAT_ERROR_PHYS < worldBox.top + worldBox.height) // The object doesn't touch the ground
        {
            Box upperBox = m_box; //Box 1px up compared to m_box
            upperBox.top += .1f;
            
            std::vector<PhysicsObject*> initialHits = hitTest();
            std::vector<PhysicsObject*> hitObjects = hitTest(upperBox);
            
            PhysicsRoutines::soustractVectors(hitObjects, initialHits);
            
            if (hitObjects.size() == 0) //The object doesn't touch another one
                return false;
            else
            {
                if (material != nullptr)
                {
                    PhysicsRoutines::sortObjectsByY(hitObjects, 0);
                    (*material) = hitObjects.front()->getMaterial();
                }
                
                return true;
            }
        }
        
        //The object is on the world's ground
        if (material != nullptr)
            (*material) = m_physicsWorld->getGroundMaterial();
        
        return true;
    }
    
    return false;
}

const Box &PhysicsObject::getBox() const
{
    return m_box;
}

std::vector<PhysicsObject*> PhysicsObject::hitTest() const
{
    return hitTest(m_box);
}

std::vector<PhysicsObject*> PhysicsObject::hitTest(const Box &box) const
{
    std::vector<PhysicsObject*> hitObjects;
    
    if (m_physicsWorld != nullptr)
        hitObjects = m_physicsWorld->hitTest((*this), box);
    
    return hitObjects;
}
