/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PlaylistPlayer.hpp"

namespace PlaylistPlayer
{
    namespace
    {
        std::map<unsigned int, PlaylistData> playlists;
        std::vector<std::thread*> threads;

        std::mutex mutex;
        
        //Launches a notification for threadID's playlist current song
        //threadID must point to a valid and running thread
        //playlists access must be secured (using mutexes) on call.
        void playlistNotify(unsigned int threadID, const Playlist &playlist)
        {
            if (playlists[threadID].notifier != nullptr && playlists[threadID].notifierMutex != nullptr)
            {
                SongData songData;
                if (playlist.getData(songData))
                {
                    playlists[threadID].notifierMutex->lock();
                    playlists[threadID].notifier->notify(songData.author, songData.title);
                    playlists[threadID].notifierMutex->unlock();
                }
            }
        }
        
        void launchThread(unsigned int threadID)
        {
            mutex.lock();
            
            if (playlists.count(threadID) == 0)
            {
                hnm::printerr(hnm::format("[Playlist] Thread {} could not be launched (no playlist data specified).",
                                          threadID) );
                mutex.unlock();
                return;
            }
            
            hnm::print( hnm::format("[Playlist] Thread {} launched for playlist #{} playback", threadID, threadID) );
            
            //Create the playlist object
            Playlist playlist;
            playlist.setVolume(0.f);
            for (const SongData &song : playlists[threadID].songs)
                playlist.addSong(song);
            
            playlist.nextSong(false);
            
            //First notification: For the start of the playlist
            playlistNotify(threadID, playlist);
            
            //Initialize the fade in
            playlists[threadID].fade_in = true;
            
            mutex.unlock();
            
            while (1)
            {
                mutex.lock();
                
                //Update the internally playing music
                playlist.update();
                
                if (playlists[threadID].volume_updated &&
                        !playlists[threadID].next_song &&
                        !playlists[threadID].fade_in)
                {
                    playlists[threadID].volume_updated = false;
                    playlist.setVolume(playlists[threadID].volume);
                }
                
                if (playlist.songAboutToEnd() && !playlists[threadID].next_song)
                {
                    hnm::print( hnm::format("[Playlist] Thread {}: switching to next song", threadID) );
                    playlists[threadID].next_song = true;
                }
                
                //Next song
                if (playlists[threadID].next_song)
                {
                    //Fade out
                    float volume = playlist.getVolume() - 1.f; //Decrease volume by 1%
                    if (volume > 0.f)
                        playlist.setVolume(volume);
                    else
                    {
                        if (!playlists[threadID].loop_next)
                            playlist.setLooping(false);
                        
                        //Start next song
                        playlist.setVolume(0.f);
                        playlist.nextSong();
                        
                        if (playlists[threadID].loop_next)
                            playlist.setLooping(true);
                        
                        //Reset the flag
                        playlists[threadID].next_song = false;
                        playlists[threadID].fade_in = true;
                        
                        //Graphics notifier handling
                        playlistNotify(threadID, playlist);
                    }
                }
                
                //Fade in
                if (playlists[threadID].fade_in)
                {
                    float volume = playlist.getVolume() + 1.f; //Increase volume by 1%
                    if (volume <= playlists[threadID].volume)
                        playlist.setVolume(volume);
                    else
                    {
                        playlist.setVolume(playlists[threadID].volume);
                        playlists[threadID].fade_in = false;
                    }
                }
                
                //Terminate
                if (playlists[threadID].stop)
                {
                    hnm::print( hnm::format("[Playlist] Exiting Thread {}", threadID) );
                    mutex.unlock();
                    
                    break;
                }
                
                mutex.unlock();
                Timer::sleep(10.f);
            }
            
            //Erase the playlist
            mutex.lock();
            playlists.erase(threadID);
            mutex.unlock();
        }
    }
    
    bool startPlaylist(const std::vector<SongData> songList, float volume, unsigned int requestedId,
                       PlaylistNotifier *notifier, std::mutex *notifierMutex)
    {
        if (!isPlaying(requestedId))
        {
            //Register data
            mutex.lock();
            playlists[requestedId] = PlaylistData(songList, volume); //requestedId was not already in playlists
            
            //Notifier handling
            if (notifier != nullptr && notifierMutex != nullptr)
            {
                playlists[requestedId].notifier = notifier; //Set the graphical notifier
                playlists[requestedId].notifierMutex = notifierMutex; //The mutex to conciliate drawing (main thread)
                                                                      //and notifying (the PlaylistPlayer thread)
            }
            
            mutex.unlock();
            
            //Create & register thread
            std::thread *thread = new std::thread(launchThread, requestedId);
            threads.push_back(thread);
            
            return true;
        }
        
        return false;
    }
    
    void stopPlaylist(unsigned int playlistId)
    {
        mutex.lock();
        
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
        {
            if (it->first == playlistId)
                it->second.stop = true; //Flag to stop the playlist
        }
        
        mutex.unlock();
    }
    
    void playlistNextSong(unsigned int playlistId, bool loop)
    {
        mutex.lock();
        
        //Search for the right playlist
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
        {
            if (it->first == playlistId)
            {
                it->second.next_song = true; //Flag to launch the next music
                it->second.loop_next = loop; //Whether to loop the next song
            }
        }
        
        mutex.unlock();
    }
    
    void playlistSetVolume(unsigned int playlistId, float musicVolume)
    {
        if (musicVolume < 0.f || musicVolume > 100.f)
            return;
        
        mutex.lock();
        
        //Search for the right playlist
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
        {
            if (it->first == playlistId)
            {
                it->second.volume = musicVolume;
                it->second.volume_updated = true; //Flag to apply the updated volume
            }
        }
        
        mutex.unlock();
    }
    
    void playlistSetNotifier(unsigned int playlistId, PlaylistNotifier *notifier, std::mutex *notifierMutex)
    {
        mutex.lock();
        
        //Search for the right playlist
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
        {
            if (it->first == playlistId)
            {
                it->second.notifier = notifier; //Set the graphical notifier
                it->second.notifierMutex = notifierMutex; //The mutex to conciliate drawing (main thread)
                                                          //and notifying (the PlaylistPlayer thread)
            }
        }
        
        mutex.unlock();
    }

    bool isPlaying(unsigned int playlistId)
    {
        mutex.lock();
        
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
        {
            if (it->first == playlistId)
            {
                mutex.unlock();
                return true;
            }
        }
        
        mutex.unlock();
        return false;
    }
    
    void destroy()
    {
        //Stop remaining threads
        mutex.lock();
        
        std::map<unsigned int, PlaylistData>::iterator it;
        for (it = playlists.begin(); it != playlists.end(); it++)
            it->second.stop = true; //Flag to stop the playlist (stop immediately)
        
        mutex.unlock();
        
        //Wait for remaining threads to exit
        for (std::thread *thread : threads)
        {
            thread->join();
            delete thread;
        }
        
        //Clear playlists & threads
        threads.clear();
        playlists.clear();
    }
}
