/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "TextMenu.hpp"

TextMenu::TextMenu() : Menu(), m_browsed(false)
{
    
}

TextMenu::TextMenu(const TextButtonSkin &skin, float space, Menu::Orientation orientation, float width, float height, const Color &color) :
Menu(space, orientation, width, height, color), m_buttonSkin(skin), m_browsed(false)
{
    
}

void TextMenu::draw()
{
    //Update matrix
    Drawable::pushMatrix();
    Drawable::updateMatrix();
    
    m_background.draw();
    
    for (TextButton &button : m_buttonsList)
        button.draw();
    
    Drawable::popMatrix();
}

void TextMenu::launch()
{
    //Clearing possible selections
    m_selections.clear();
    m_browsed = false;
    
    //Drawing root elements
    Menu::resetCurrentElement();
    draw(m_currentElement);
}

bool TextMenu::mouseRelease(const Vector &mouseCoords)
{
    Vector localMouse = Widget::localCoords(mouseCoords);
    
    //Browse the list
    for (TextButton &button : m_buttonsList)
    {
        if (button.mouseRelease(localMouse)) //Button clicked
        {
            std::string buttonString = button.getText();
            
            //Search the matching TinyXML element
            pugi::xml_node itemsNode;
            for (itemsNode = m_currentElement;
                 itemsNode;
                 itemsNode = itemsNode.next_sibling("item"))
            {
                std::string itemName(itemsNode.attribute("content").value());
                std::string itemId(itemsNode.attribute("id").value());
                
                if (itemName == buttonString) //Found
                {
                    //Register selection in m_selections
                    m_selections.push_back(itemId);
                    
                    if (!itemsNode.child("item")) //Selection ended
                        m_browsed = true;
                    else //Continue browsing
                    {
                        //Draw next items
                        m_currentElement = itemsNode.child("item");
                        draw(m_currentElement);
                    }
                    
                    return true; //End the fonction now
                }
            }
        }
    }
    
    return false;
}

bool TextMenu::mouseClick(const Vector &mouseCoords)
{
    Vector localMouse = Widget::localCoords(mouseCoords);
    return Widget::contains(localMouse);
}

bool TextMenu::mouseMove(const Vector &mouseCoords)
{
    Vector localMouse = Widget::localCoords(mouseCoords);
    
    for (TextButton &button : m_buttonsList)
        button.mouseMove(localMouse);
    
    return Widget::contains(localMouse);
}

bool TextMenu::browsed() const
{
    return m_browsed;
}

const std::vector<std::string> &TextMenu::result() const
{
    return m_selections;
}

void TextMenu::draw(pugi::xml_node itemsList)
{
    m_buttonsList.clear();
    
    int i = 0;
    while (itemsList)
    {
        std::string itemName(itemsList.attribute("content").value());
        TextButton itemButton(itemName, m_buttonSkin);
        
        if (m_orientation == Menu::Orientation::Vertical)
            itemButton.setPosition( (m_width - contentWidth()) / 2.f,
                                    20.f + static_cast<float>(i) * (itemButton.getHeight() + m_space) );
        else if (m_orientation == Menu::Orientation::Horizontal)
            itemButton.setPosition( 20.f + static_cast<float>(i) * (itemButton.getWidth() + m_space),
                                    (m_height - contentHeight()) / 2.f );
        
        m_buttonsList.push_back(itemButton);
        
        itemsList = itemsList.next_sibling("item");
        i++;
    }
}

float TextMenu::contentWidth() const
{
    float width = 0.f;
    
    if (m_orientation == Menu::Orientation::Vertical)
        width = m_buttonSkin.skin.box.width;
    
    else if (m_orientation == Menu::Orientation::Horizontal)
        width = (m_buttonSkin.skin.box.width + m_space) * static_cast<float>(m_buttonsList.size());
    
    return width;
}

float TextMenu::contentHeight() const
{
    float height = 0.f;
    
    if (m_orientation == Menu::Orientation::Vertical)
        height = (m_buttonSkin.skin.box.height + m_space) * static_cast<float>(m_buttonsList.size());
    
    else if (m_orientation == Menu::Orientation::Horizontal) //Menu is horizontal
        height = m_buttonSkin.skin.box.height;
    
    return height;
}

//Get and set methods
void TextMenu::setSkin(const TextButtonSkin &skin)
{
    m_buttonSkin = skin;
    
    if (m_currentElement)
        draw(m_currentElement);
}
