/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "AI.hpp"

AI::AI() : AI("", 100.f, 0.f, 0.f, 0.f, 0.f) {}

AI::AI(const std::string &name, float health, float strength, float speed, float jumpEnergy, float walkPeriod) :
Player(name, health, strength, speed, jumpEnergy, walkPeriod),
m_generator( static_cast<std::mt19937::result_type>( Timer::getSystemTime().count() ) ),
m_actionDist(0, 999), m_targetEnemy(0), m_behaviour(AI::Behaviour::Offensive), m_attackState_start(0.f),
m_attackPause(false), m_attackPause_start(0.f), m_defensive_turnaway(false), m_defensive_turnaway_start(0.f)
{
    
}

void AI::act()
{
    //Handling KO enemies (not to focus on a poor ko player ;))
    updateEnemies();
    
    if (m_moves.getAction() == ActionKey::Default && !isKo() && !isHit() && m_enemies.size() == 0) //No more enemies so stop !
    {
        setMoving(Player::MovingState::NotMoving);
        if (m_moves.getState() == StateKey::Crouched)
            m_moves.setState(StateKey::Default);
    }
    else if (m_moves.getAction() == ActionKey::Default && !isKo() && !isHit() && m_enemies.size() > 0) //Main AI Behaviour block
    {
        //Handling target enemy
        if (m_targetEnemy >= m_enemies.size() || m_actionDist(m_generator) < 10) //Bad index, or a 1% chance to randomly change target
        {
            std::uniform_int_distribution<unsigned int> enemyDist(0, static_cast<unsigned int>(m_enemies.size())-1);
            m_targetEnemy = enemyDist(m_generator);
        }
        
        //Get information on the target enemy's state
        const Player &enemy = (*m_enemies[m_targetEnemy]);
        
        /*unsigned int e_state = 0, e_action = 0;
        enemy.getState(e_state, e_action);*/
        
        //Define the behaviour of the AI
        if (enemy.getHealth() <= 25.f)
            m_behaviour = AI::Behaviour::Offensive;
        else if (getHealth() <= 30.f)
            m_behaviour = AI::Behaviour::Defensive;
        
        //Implement the different behaviours of the AI
        if (m_behaviour == AI::Behaviour::Defensive)
        {
            Box enemyBox = enemy.getBox();
            
            //If the enemy is too close
            if (enemyBox.left - 300.f <= getBox().left + getBox().width && enemyBox.left + enemyBox.width + 300.f >= getBox().left)
            {
                Box worldBox = getPhysicsWorld().getWorldBox();
                
                //Move away if not too close to the world limits
                float world_right = worldBox.left + worldBox.width;
                float ai_right = getBox().left + getBox().width;
                
                if (world_right - ai_right >= 200.f && getBox().left - worldBox.left >= 200.f)
                {
                    setXFlip(enemyBox.left >= getBox().left);
                    setMoving(getAheadDirection());
                    
                    if (m_actionDist(m_generator) < 35) jump();
                }
                else //Or fight back
                {
                    //Run away if possible
                    bool enemy_left = enemyBox.left <= getBox().left; //Is the enemy to the left ?
                    
                    if ( (world_right - ai_right < 200.f && !enemy_left) || (getBox().left - worldBox.left < 200.f && enemy_left) )
                    {
                        setXFlip(!enemy_left); //Turn away
                        setMoving(getAheadDirection());
                    }
                    else
                    {
                        setMoving(Player::MovingState::NotMoving);
                        setXFlip(enemyBox.left < getBox().left);
                        
                        int action = m_actionDist(m_generator);
                        if (action < 50) punch();
                        else if (action < 100) kick();
                        else if (action < 125) jump();
                    }
                }
            }
            else if (enemyBox.left - 450.f > getBox().left + getBox().width || enemyBox.left + enemyBox.width + 450.f < getBox().left)
            {
                //If the distance between the AI and its enemy is fair enough
                randomDefensive();
            }
        }
        else if (m_behaviour == AI::Behaviour::Offensive)
        {
            Box enemyBox = enemy.getBox();
            
            //Check if the enemy is on ground above or the opposite, then crouch
            bool aboveEnemy = isOnGround() && getBox().top + getBox().height < enemyBox.top + enemyBox.height;
            bool enemyAbove = enemy.isOnGround() && getBox().top + getBox().height > enemyBox.top + enemyBox.height;
            
            if (aboveEnemy || enemyAbove)
            {
                setMoving(getAheadDirection());
                if (m_actionDist(m_generator) < 10) jump();
            }
            else
            {
                //X Flip to face the enemy
                if (enemyBox.left - 100.f > getBox().left || enemyBox.left + 100.f < getBox().left)
                    setXFlip( enemyBox.left < getBox().left );
                
                //If the enemy is too far away, run to him
                if (enemyBox.left - 30.f > getBox().left + getBox().width || enemyBox.left + enemyBox.width + 30.f < getBox().left)
                {
                    //Attack pauses handling ! (sometimes make a break)
                    if (m_attackPause && Timer::getTicks() >= m_attackPause_start + 300.f)
                        m_attackPause = false;
                    
                    if (!m_attackPause && m_actionDist(m_generator) < 40)
                    {
                        setMoving(Player::MovingState::NotMoving);
                        
                        int crouch_or_jump = m_actionDist(m_generator);
                        if (crouch_or_jump < 125 && m_moves.getState() != StateKey::Crouched)
                            crouch();
                        else if (crouch_or_jump < 190)
                            jump();
                        
                        m_attackPause = true;
                        m_attackPause_start = Timer::getTicks();
                    }
                    
                    if (!m_attackPause) //Run to the enemy if no pause
                        setMoving(getAheadDirection());
                }
                else //Fight the enemy
                {
                    int action = m_actionDist(m_generator);
                    if (action < 50) punch();
                    else if (action < 100) kick();
                    else if (action < 105) jump();
                    else if (m_attackState_start + 300.f <= Timer::getTicks())
                    {
                        setMoving(Player::MovingState::NotMoving);
                        int actionState = m_actionDist(m_generator);
                        
                        if (actionState < 700 && m_moves.getState() != StateKey::Default) m_moves.setState(StateKey::Default);
                        else if (actionState < 870 && m_moves.getState() != StateKey::Crouched) crouch();
                        else setMoving(Player::MovingState::NotMoving);
                        
                        m_attackState_start = Timer::getTicks();
                    }
                }
            }
        }
    }
}

void AI::randomDefensive()
{
    //Don't move for too long
    if (isMoving())
    {
        int actionStopMove = m_actionDist(m_generator);
        if (actionStopMove < 30)
            setMoving(Player::MovingState::NotMoving);
    }
    
    if (m_actionDist(m_generator) < 670) //Sometimes just do nothing
    {
        if (m_enemies.size() == 1)
        {
            //Turn away sometimes
            if (m_defensive_turnaway && Timer::getTicks() >= m_defensive_turnaway_start + 300.f)
                m_defensive_turnaway = false;
            
            if (!m_defensive_turnaway && m_actionDist(m_generator) < 10)
            {
                setXFlip( m_enemies[0]->getBox().left >= getBox().left );
                
                m_defensive_turnaway = true;
                m_defensive_turnaway_start = Timer::getTicks();
            }
            
            if (!m_defensive_turnaway)
                setXFlip( m_enemies[0]->getBox().left < getBox().left );
        }
        else
        {
            //Flip randomly
            int actionFlip = m_actionDist(m_generator);
            if (actionFlip < 7)
                setXFlip( !m_moves.getXFlip() );
        }
        
        //Act randomly
        int action = m_actionDist(m_generator);
        setMoving(Player::MovingState::NotMoving);
        
        if (action < 40) setMoving(getAheadDirection());
        else if (action < 48) jump();
        else if (action < 56) punch();
        else if (action < 64) kick();
        else if (action < 72) throwShuriken();
    }
}

void AI::updateEnemies()
{
    //Remove all KO enemies from m_enemies
    std::vector<const Player*>::iterator it = m_enemies.begin();
    while (it != m_enemies.end())
    {
        if ((*it)->isKo())
            it = m_enemies.erase(it);
        else
            it++;
    }
}

void AI::takeAHit(float strength, HitDirection side, const PhysicsObject *source)
{
    Player::takeAHit(strength, side, source);
    
    //The AI knows who hit it
    if (source != nullptr)
    {
        //See if it's a know enemy and if so, focus it (likely)
        for (unsigned int i = 0; i < m_enemies.size(); i++)
        {
            //Found the enemy
            if (static_cast<const PhysicsObject*>(m_enemies[i]) == source)
            {
                if (m_actionDist(m_generator) < 800) //80% probability to focus it
                    m_targetEnemy = i;
            }
        }
    }
}

void AI::addEnemy(const Player &enemy)
{
    m_enemies.push_back(&enemy);
    
    std::uniform_int_distribution<unsigned int> enemyDist(0, static_cast<unsigned int>(m_enemies.size())-1);
    m_targetEnemy = enemyDist(m_generator);
}

void AI::clearEnemies()
{
    m_enemies.clear();
}

const std::vector<const Player*> &AI::getEnemies() const
{
    return m_enemies;
}

//Static methods
AI AI::loadFromXML(const std::string &name, const std::string &xmlRelPath)
{
    //TinyXML initialization
    std::string xmlPath = BuildValues::data(xmlRelPath); //Locate absolute path to xml data
    
    pugi::xml_document xmlFile;
    if (!xmlFile.load_file(xmlPath.c_str()))
        return AI();
    
    pugi::xml_node infoNode;
    for (infoNode = xmlFile.child("main").child("character");
         infoNode;
         infoNode = infoNode.next_sibling("character"))
    {
        std::string itemName(infoNode.attribute("name").value());
        if (itemName == name)
        {
            //Get information
            float strength = infoNode.attribute("strength").as_float(10.f),
                  speed = infoNode.attribute("speed").as_float(700.f),
                  jumpEnergy = infoNode.attribute("jumpEnergy").as_float(750.f),
                  walkPeriod = infoNode.attribute("walkPeriod").as_float(.2f);
            
            //Create player
            return AI(name, 100.f, strength, speed, jumpEnergy, walkPeriod);
        }
    }
    
    return AI();
}
