/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Player.hpp"

Player::Player() : Player("", 100.f, 0.f, 0.f, 0.f, 0.f) {}

Player::Player(const std::string &name, float health, float strength, float speed, float jumpEnergy, float walkPeriod) :
PhysicsObject(), m_name(name), m_health(health), m_strength(strength), m_speed(speed), m_jumpEnergy(jumpEnergy), m_walkPeriod(walkPeriod),
m_showHitbox(false), m_soundEngine(nullptr), m_weapons(nullptr), m_lockedEnnemy(nullptr), m_requestsMove(Player::MovingState::NotMoving),
m_awaitsImpact(false), m_debugHits(false)
{
    
}

void Player::initWeapons(WeaponMedium &medium)
{
    m_weapons = &medium;
}

void Player::loadGraphics(TextureManager &textureManager)
{
    std::string graphicsPath = "gfx/characters/" + m_name + "/";
    
    m_moves.lockToGrid(false);
    m_moves.load(graphicsPath + "data.xml", graphicsPath, textureManager, (*this));
    m_moves.update(0); //Initial update (origin, ...)
    
    //Init physics box
    setWidth(m_moves.getWidth());
    setHeight(m_moves.getHeight());
    
    //Init shadow
    float shadowRatio = .1f;
    
    m_shadowSize.x = m_moves.getWidth() * 1.1f;
    m_shadowSize.y = m_shadowSize.x * shadowRatio;
    
    m_shadow = Polygon::circle(m_shadowSize.x / 2.f);
    m_shadow.lockToGrid(false);
    m_shadow.setBorderSize(0.f);
    m_shadow.setYScale(shadowRatio);
}

void Player::draw()
{
    if (!m_moves.getXFlip()) m_moves.setXPosition(m_box.left);
    else m_moves.setXPosition(m_box.left + m_box.width);
    
    m_moves.setYPosition(m_box.top + m_box.height);
    
    float shadowDist = shadowProjection();
    if (shadowDist < 0.f) shadowDist = 0.f;
    
    m_shadow.setPosition( m_box.left - (m_shadowSize.x - m_box.width) / 2.f,
                          m_box.top + m_box.height + shadowDist - m_shadowSize.y / 2.f );
    m_shadow.setUniformColor( Color( 0, 0, 0, static_cast<GLubyte>(90.f / (shadowDist/50.f + 1.f)) ) );
    
    m_shadow.draw();
    m_moves.draw();
    
    if (m_showHitbox)
    {
        m_hitboxView.draw();
        m_showHitbox = false;
    }
}

void Player::update(float frameTime)
{
    m_moves.update(frameTime);
    updatePhysicsBox();
    
    if (!isKo() && !isHit())
    {
        std::string groundMaterial;
        bool onGround = isOnGround(&groundMaterial);
        
        //Impact sounds
        if (!onGround) m_awaitsImpact = true;
        else if (onGround && m_awaitsImpact)
        {
            m_awaitsImpact = false;
            if (groundMaterial != MATERIAL_NONE)
                playSoundEffect("impact_" + groundMaterial);
        }
        
        //Jumping
        if (onGround && m_moves.getState() == StateKey::Jumping)
        {
            m_moves.setState(StateKey::Default);
            updatePhysicsBox();
        }
        else if (!onGround)
        {
            m_moves.setState(StateKey::Jumping);
            updatePhysicsBox();
        }
        
        //Moving
        if (m_requestsMove != Player::MovingState::NotMoving)
        {
            if (isOnGround() && m_moves.getState() != StateKey::Moving)
            {
                m_moves.setState(StateKey::Moving);
                updatePhysicsBox();
            }
            
            moveStep(frameTime);
        }
        
        //Ennemy locking
        if (m_lockedEnnemy != nullptr && m_moves.getState() != StateKey::Moving)
            m_moves.setXFlip( (m_lockedEnnemy->getBox().left < getBox().left) );
    }
}

void Player::updatePhysicsBox()
{
    setWidth(m_moves.getWidth(), m_moves.getXFlip());
    setHeight(m_moves.getHeight(), true);
}

void Player::lockEnnemy(const Player *ennemy)
{
    m_lockedEnnemy = ennemy;
}

void Player::moveStep(float frameTime)
{
    if (!isKo() && !isHit()) //Move
    {
        float elapsedTime = frameTime / 1000.f;
        
        float speed = m_speed;
        if (movingAwayFromEnnemy()) speed /= 2.f;
        
        if (m_requestsMove == Player::MovingState::MovingLeft) moveXPosition(-speed * elapsedTime);
        else if (m_requestsMove == Player::MovingState::MovingRight) moveXPosition(speed * elapsedTime);
    }
}

void Player::setMoving(Player::MovingState direction)
{
    if (m_moves.getAction() == ActionKey::Default)
    {
        m_requestsMove = direction;
        
        // Update player state immediately.
        if (m_requestsMove == Player::MovingState::NotMoving && m_moves.getState() == StateKey::Moving)
        {
            m_moves.setState(StateKey::Default);
            updatePhysicsBox();
        }
        else if (m_requestsMove != Player::MovingState::NotMoving && isOnGround() && m_moves.getState() != StateKey::Moving)
        {
            m_moves.setState(StateKey::Moving);
            updatePhysicsBox();
        }
    }
}

void Player::jump()
{
    if (m_physicsWorld != nullptr && (!isKo() && !isHit() && m_moves.getState() != StateKey::Crouched && isOnGround()))
        m_physicsWorld->pulseObject((*this), 0.f, -m_jumpEnergy);
}

void Player::crouch()
{
    if (!isKo() && !isHit() && isOnGround())
        m_moves.setState(StateKey::Crouched);
}

void Player::punch()
{
    if (!isKo() && !isHit() && m_moves.mayAttack() && !movingAwayFromEnnemy())
        m_moves.punch();
}

void Player::kick()
{
    if (!isKo() && !isHit() && m_moves.mayAttack() && !movingAwayFromEnnemy())
        m_moves.kick();
}

void Player::throwShuriken()
{
    if (m_weapons != nullptr && m_physicsWorld != nullptr &&
        !isKo() && !isHit() && m_moves.mayAttack() && !movingAwayFromEnnemy())
    {
        m_weapons->throwShuriken((*m_physicsWorld), (*this));
    }
}

void Player::collideWorld(bool /*ground*/)
{
    //Nothing to do :)
}

void Player::collide(PhysicsObject& /*obj*/)
{
    //Nothing to do :)
}

void Player::takeAHit(float strength, HitDirection side, const PhysicsObject* /*source*/)
{
    //Alter life
    float remainingLives = m_health - strength;
    setHealth(remainingLives);
    
    //Pulse
    float pulse = 40.f * strength;
    if (side == HitDirection::FromRight) pulse *= -1.f;
    
    if (m_physicsWorld != nullptr)
        m_physicsWorld->pulseObject((*this), pulse, 0.f);
     
    //Launch KO animation
    if (!isKo() && remainingLives <= 0.f)
        m_moves.setState(StateKey::Falling, StateKey::KO, true);
    else if (!isKo() && !isHit()) //Or launch hit animation
        m_moves.takeAHit();
}

bool Player::isKo() const
{
    return (m_moves.getState() == StateKey::Falling || m_moves.getState() == StateKey::KO);
}

bool Player::isHit() const
{
    return (m_moves.getAction() == ActionKey::TakeAHit);
}

bool Player::isMoving() const
{
    return (m_requestsMove != Player::MovingState::NotMoving);
}

bool Player::movingAwayFromEnnemy() const
{
    return (m_lockedEnnemy != nullptr && (
            (getAheadDirection() == Player::MovingState::MovingLeft && m_lockedEnnemy->getBox().left >= getBox().left) ||
            (getAheadDirection() == Player::MovingState::MovingRight && m_lockedEnnemy->getBox().left < getBox().left)));
}

Player::MovingState Player::getAheadDirection() const
{
    if (m_moves.getXFlip())
        return Player::MovingState::MovingLeft;
    
    return Player::MovingState::MovingRight;
}

//Attacks callback
void Player::hit(const HitBox &hitBox)
{
    //Define the hit box
    Box relativeBox = hitBox.box;
    relativeBox.left = m_box.left + relativeBox.left;
    relativeBox.top = m_box.top + relativeBox.top;
    
    if (m_moves.getXFlip())
        relativeBox.left = m_box.left - ((relativeBox.left + relativeBox.width) - (m_box.left + m_box.width));
    
    //Draw a red square around it (if needed)
    if (m_debugHits)
    {
        m_hitboxView = Polygon::rectangle(relativeBox.left,
                                            relativeBox.top,
                                            relativeBox.width,
                                            relativeBox.height,
                                            Color(0, 0, 0, 0),
                                            Color(225, 0, 0));
        m_hitboxView.setBorderSize(4.f);
        m_showHitbox = true;
    }
    
    //Hit objects
    std::vector<PhysicsObject*> hitObjects = hitTest(relativeBox);
    
    for (unsigned int i = 0; i < hitObjects.size(); i++) //Pum ! ;)
    {
        if (m_box.left < hitObjects[i]->getBox().left)
            hitObjects[i]->takeAHit(m_strength * hitBox.factor, HitDirection::FromLeft, this);
        else
            hitObjects[i]->takeAHit(m_strength * hitBox.factor, HitDirection::FromRight, this);
    }
}

void Player::playSoundEffect(const std::string &effectName)
{
    if (m_soundEngine != nullptr)
        m_soundEngine->playSoundEffect(effectName, m_name);
}

//Animations callback
void Player::stepImpactCallback()
{
    std::string stepMaterial = MATERIAL_NONE;
    if (isOnGround(&stepMaterial))
    {
        if (stepMaterial != MATERIAL_NONE)
            playSoundEffect("walk_" + stepMaterial);
    }
}

//Engines methods
void Player::setSoundEngine(SoundEngine &soundEngine)
{
    m_soundEngine = &soundEngine;
}

//Get and set methods
const std::string &Player::getName() const
{
    return m_name;
}

void Player::setHealth(float health)
{
    if (health < 0.f)
        m_health = 0.f;
    
    else if (health > 100.f)
        m_health = 100.f;
    
    else
        m_health = health;
}

float Player::getHealth() const
{
    return m_health;
}

void Player::setStrength(float strength)
{
    if (strength >= 0.f)
        m_strength = strength;
}

float Player::getStrength() const
{
    return m_strength;
}

void Player::setSpeed(float speed)
{
    if (speed >= 0.f)
        m_speed = speed;
}

float Player::getSpeed() const
{
    return m_speed;
}

void Player::setJumpEnergy(float jumpEnergy)
{
    if (jumpEnergy >= 0.f)
        m_jumpEnergy = jumpEnergy;
}

float Player::getJumpEnergy() const
{
    return m_jumpEnergy;
}

void Player::setDebugHits(bool debugHits)
{
    m_debugHits = debugHits;
}

bool Player::getDebugHits() const
{
    return m_debugHits;
}

void Player::setXFlip(bool flip)
{
    m_moves.setXFlip(flip);
}

bool Player::getXFlip() const
{
    return m_moves.getXFlip();
}

bool Player::getWeapons(WeaponSlots &dest) const
{
    if (m_weapons != nullptr)
        return m_weapons->getSlots((*this), dest);
    
    return false;
}
