/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PlayerMoves.hpp"
#include "Player.hpp"

PlayerMoves::PlayerMoves() : Drawable(), m_state(StateKey::Default), m_action(ActionKey::Default),
m_targetState(StateKey::None), m_lastAttackEndTime(-RECOVERY_TIME - 1.f)
{
    
}

void PlayerMoves::draw()
{
    //Update matrix
    Drawable::pushMatrix();
    Drawable::updateMatrix();
    
    if (m_action != ActionKey::Default) m_statesList[m_state].attacks[m_action].draw();
    else m_statesList[m_state].animation.draw();
    
    Drawable::popMatrix();
}

void PlayerMoves::update(float frameTime)
{
    if (m_action != ActionKey::Default) //Attack launched
    {
        m_statesList[m_state].attacks[m_action].update(frameTime);
        
        if (m_statesList[m_state].attacks[m_action].ready()) //finished
        {
            m_action = ActionKey::Default;
            
            m_statesList[m_state].animation.goToTime(0.f);
            updateOrigin();
            
            m_lastAttackEndTime = Timer::getTicks();
            return;
        }
    }
    else
    {
        CallbackAnimation &animation = m_statesList[m_state].animation;
        if (animation.atEndedState())
        {
            animation.goToTime(0.f);
            
            if (m_targetState != StateKey::None)
            {
                setState(m_targetState);
                m_targetState = StateKey::None;
                
                return;
            }
        }
        
        animation.goForward(frameTime / 1000.f);
    }
    
    updateOrigin();
}

//Action
bool PlayerMoves::mayAttack()
{    
    return m_action == ActionKey::Default && (Timer::getTicks() - m_lastAttackEndTime) >= RECOVERY_TIME;
}

void PlayerMoves::punch()
{
    if (mayAttack() && m_statesList[m_state].attacks.count(ActionKey::Punch) == 1)
    {
        m_action = ActionKey::Punch;
        updateOrigin();
    }
}

void PlayerMoves::kick()
{
    if (mayAttack() && m_statesList[m_state].attacks.count(ActionKey::Kick) == 1)
    {
        m_action = ActionKey::Kick;
        updateOrigin();
    }
}

void PlayerMoves::takeAHit()
{
    if (m_statesList[m_state].attacks.count(ActionKey::TakeAHit) == 1)
    {
        //Reinit current attack if needed
        if (m_action != ActionKey::Default)
            m_statesList[m_state].attacks[m_action].reinit();
        
        m_action = ActionKey::TakeAHit;
        updateOrigin();
    }
}

//State
void PlayerMoves::setState(StateKey state, bool force)
{
    //Forcing to change state
    //(basically in case of KO : current attacks must not interfere)
    if (force && m_action != ActionKey::Default)
    {
        //Reinit current attack and properly go back to default action
        m_statesList[m_state].attacks[m_action].reinit();
        m_action = ActionKey::Default;
    }
    
    if (m_action == ActionKey::Default)
    {
        m_state = state;
        
        m_statesList[m_state].animation.goToTime(0.f);
        updateOrigin();
    }
}

void PlayerMoves::setState(StateKey state, StateKey targetState, bool force)
{
    if (m_action == ActionKey::Default || force)
    {
        setState(state, force);
        m_targetState = targetState;
    }
}

ActionKey PlayerMoves::getAction() const
{
    return m_action;
}

StateKey PlayerMoves::getState() const
{
    return m_state;
}

float PlayerMoves::getWidth()
{
    const CallbackAnimation &anim = getAnimation();
    
    Box bodyBox;
    anim.getBodyBox(bodyBox);
    
    return bodyBox.width * anim.getXScale();
}

float PlayerMoves::getHeight()
{
    const CallbackAnimation &anim = getAnimation();
    
    Box bodyBox;
    anim.getBodyBox(bodyBox);
    
    return bodyBox.height * anim.getYScale();
}

const CallbackAnimation &PlayerMoves::getAnimation()
{
    if (m_action != ActionKey::Default)
        return m_statesList[m_state].attacks[m_action].getAnimation();
    
    return m_statesList[m_state].animation;
}

void PlayerMoves::updateOrigin()
{
    const CallbackAnimation &anim = getAnimation();
    
    Box bodyBox;
    anim.getBodyBox(bodyBox);
    
    setXOrigin(bodyBox.left * anim.getXScale());
    setYOrigin((bodyBox.top + bodyBox.height) * anim.getYScale());
}

void PlayerMoves::load(const std::string &xmlRelPath, const std::string &animationsPath, TextureManager &textureManager, Player &player)
{
    //TinyXML initialization
    std::string xmlPath = BuildValues::data(xmlRelPath); //Locate absolute path to xml data
    
    pugi::xml_document xmlFile;
    if (!xmlFile.load_file(xmlPath.c_str()))
        return;
    
    pugi::xml_node stateNode;
    for (stateNode = xmlFile.child("main").child("data").first_child();
         stateNode;
         stateNode = stateNode.next_sibling())
    {
        std::string stateName(stateNode.name());
        StateKey stateKey;
        if (stateName == "default") stateKey = StateKey::Default;
        else if (stateName == "jumpTransition") stateKey = StateKey::JumpTransition;
        else if (stateName == "jumping") stateKey = StateKey::Jumping;
        else if (stateName == "moving") stateKey = StateKey::Moving;
        else if (stateName == "crouched") stateKey = StateKey::Crouched;
        else if (stateName == "falling") stateKey = StateKey::Falling;
        else if (stateName == "ko") stateKey = StateKey::KO;
        else continue;
        
        State state;
        
        //Main animation
        pugi::xml_node animNode = stateNode.child("animation");
        
        std::string animationPath = animationsPath + std::string(animNode.attribute("path").value()) + ".png";
        Texture *texture = textureManager.getTexture(animationPath);
        
        loadAnimation((*texture), animNode, state.animation);
        state.animation.setReceiver(&player);
        
        //Attacks
        pugi::xml_node attackNode;
        for (attackNode = stateNode.child("attacks").first_child();
             attackNode;
             attackNode = attackNode.next_sibling())
        {
            std::string attackName(attackNode.name());
            ActionKey actionKey;
            if (attackName == "punch") actionKey = ActionKey::Punch;
            else if (attackName == "kick") actionKey = ActionKey::Kick;
            else if (attackName == "takeAHit") actionKey = ActionKey::TakeAHit;
            else continue;
            
            Attack &attack = state.attacks[actionKey];
            attack.setPlayer(player);
            
            //Animation
            pugi::xml_node attackAnimNode = attackNode.child("animation");
            
            CallbackAnimation animation;
            animation.setReceiver(&player);
            
            std::string animationPath = animationsPath + std::string(attackAnimNode.attribute("path").value()) + ".png";
            Texture *texture = textureManager.getTexture(animationPath);
            
            loadAnimation((*texture), attackAnimNode, animation);
            attack.setAnimation(animation);
            
            //Hits
            pugi::xml_node hitNode;
            for (hitNode = attackNode.child("hits").child("hit");
                 hitNode;
                 hitNode = hitNode.next_sibling("hit"))
            {
                //Box
                Box hitBox;
                float x = hitNode.attribute("x").as_float(0.f),
                      y = hitNode.attribute("y").as_float(0.f),
                      width = hitNode.attribute("width").as_float(-1.f),
                      height = hitNode.attribute("height").as_float(-1.f);
                
                if (width < 0.f || height < 0.f)
                    hitBox = Box(x - 5.f, y - 5.f, 10.f, 10.f);
                else
                    hitBox = Box(x, y, width, height);
                
                //Time & Strength factor
                float hitTime = hitNode.attribute("time").as_float(0.f),
                      hitFactor = hitNode.attribute("strength").as_float(0.f);
                
                attack.addHitBox( HitBox(hitBox, hitTime, hitFactor) );
            }
            
            //Scale the hit boxes according to the animation scale
            attack.scaleHitBoxes(ANIMATION_SCALE);
            
            //Sound effects
            pugi::xml_node soundeffectNode;
            for (soundeffectNode = attackNode.child("sfx").child("soundeffect");
                 soundeffectNode;
                 soundeffectNode = soundeffectNode.next_sibling("soundeffect"))
            {
                float effectTime = soundeffectNode.attribute("time").as_float(0.f);
                std::string effectName(soundeffectNode.attribute("name").as_string(""));
                
                attack.addSoundEffect(effectTime, effectName);
            }
        }
        
        //Add state
        m_statesList[stateKey] = state;
    }
}

void PlayerMoves::loadAnimation(Texture &texture, const pugi::xml_node &animNode, CallbackAnimation &dst) const
{
    //Main parameters
    float duration = animNode.attribute("duration").as_float(0.f);
    
    dst.setTexture(texture);
    dst.setDuration(duration);
    dst.setScale(ANIMATION_SCALE);
    
    //Frames
    pugi::xml_node frameNode;
    for (frameNode = animNode.child("frame");
         frameNode;
         frameNode = frameNode.next_sibling("frame"))
    {
        int srcX = frameNode.attribute("x").as_int(0),
            srcY = frameNode.attribute("y").as_int(0),
            srcWidth = frameNode.attribute("width").as_int(0),
            srcHeight = frameNode.attribute("height").as_int(0),
            repeat = frameNode.attribute("repeat").as_int(1);
        
        float bodyX = frameNode.attribute("bodyX").as_float(0.f),
              bodyY = frameNode.attribute("bodyY").as_float(0.f),
              bodyWidth = frameNode.attribute("bodyWidth").as_float(0.f),
              bodyHeight = frameNode.attribute("bodyHeight").as_float(0.f);
        
        bool stepImpact = frameNode.attribute("stepImpact").as_bool(false);
        
        Frame frame( Box(static_cast<float>(srcX),
                         static_cast<float>(srcY),
                         static_cast<float>(srcWidth),
                         static_cast<float>(srcHeight)), Box(bodyX, bodyY, bodyWidth, bodyHeight) );
        
        for (int i = 0; i < repeat; i++)
            dst.addFrame(frame);
        
        if (stepImpact)
            dst.addStepImpactCallback(dst.getFrameSize()-1);
    }
}
