/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "CharacterPreview.hpp"

CharacterPreview::CharacterPreview() : Drawable(), m_selected(false)
{
    
}

void CharacterPreview::draw()
{
    //Update matrix
    Drawable::pushMatrix();
    Drawable::updateMatrix();
    
    if (m_selected) m_selectedAnim.draw();
    else m_defaultAnim.draw();
    
    Drawable::popMatrix();
}

void CharacterPreview::update(float frameTime)
{
    Animation &anim = m_selected ? m_selectedAnim : m_defaultAnim;
    
    if (anim.atEndedState())
        anim.goToTime(0.f);
    
    anim.goForward(frameTime / 1000.f);
}

void CharacterPreview::load(const std::string &characterName, TextureManager &manager)
{
    loadAnimations(characterName, m_defaultAnim, m_selectedAnim, manager);
}

void CharacterPreview::select()
{
    m_selected = true;
}

void CharacterPreview::unselect()
{
    m_selected = false;
}

float CharacterPreview::getWidth() const
{
    const Animation &anim = m_selected ? m_selectedAnim : m_defaultAnim;
    return anim.getWidth();
}

float CharacterPreview::getBodyWidth() const
{
    const Animation &anim = m_selected ? m_selectedAnim : m_defaultAnim;
    
    Box bodyBox;
    if (!anim.getBodyBox(bodyBox))
        return 0.f;
    
    return bodyBox.width;
}

float CharacterPreview::getHeight() const
{
    const Animation &anim = m_selected ? m_selectedAnim : m_defaultAnim;
    return anim.getHeight();
}

float CharacterPreview::getBodyHeight() const
{
    const Animation &anim = m_selected ? m_selectedAnim : m_defaultAnim;
    
    Box bodyBox;
    if (!anim.getBodyBox(bodyBox))
        return 0.f;
    
    return bodyBox.height;
}

void CharacterPreview::loadAnimation(Texture &texture, const pugi::xml_node &animNode, Animation &dst) const
{
    //Main parameters
    float duration = animNode.attribute("duration").as_float(0.f);
    
    dst.setTexture(texture);
    dst.setDuration(duration);
    
    //Frames
    pugi::xml_node frameNode;
    for (frameNode = animNode.child("frame");
         frameNode;
         frameNode = frameNode.next_sibling("frame"))
    {
        int srcX = frameNode.attribute("x").as_int(0),
            srcY = frameNode.attribute("y").as_int(0),
            srcWidth = frameNode.attribute("width").as_int(0),
            srcHeight = frameNode.attribute("height").as_int(0),
            repeat = frameNode.attribute("repeat").as_int(1);
        
        float bodyX = frameNode.attribute("bodyX").as_float(0.f),
              bodyY = frameNode.attribute("bodyY").as_float(0.f),
              bodyWidth = frameNode.attribute("bodyWidth").as_float(0.f),
              bodyHeight = frameNode.attribute("bodyHeight").as_float(0.f);
        
        Frame frame( Box(static_cast<float>(srcX),
                         static_cast<float>(srcY),
                         static_cast<float>(srcWidth),
                         static_cast<float>(srcHeight)), Box(bodyX, bodyY, bodyWidth, bodyHeight) );
        
        for (int i = 0; i < repeat; i++)
            dst.addFrame(frame);
    }
}

void CharacterPreview::loadAnimations(const std::string &player,
                                        Animation &dstDefault,
                                        Animation &dstSelected,
                                        TextureManager &manager) const
{
    //Locate absolute path to xml data
    std::string xmlPath = BuildValues::data( "gfx/characters/" + player + "/data.xml" );
    std::string animationsRelPath = "gfx/characters/" + player + "/";
    
    //TinyXML initialization
    pugi::xml_document xmlFile;
    if (!xmlFile.load_file(xmlPath.c_str()))
        return;
    
    pugi::xml_node baseNode = xmlFile.child("main").child("data");
    pugi::xml_node defaultNode = baseNode.child("default").child("animation"),
                   crouchedNode = baseNode.child("crouched").child("animation");
    
    //Default state -> default animation
    Texture *textureDefault = manager.getTexture( animationsRelPath +
                                                 std::string(defaultNode.attribute("path").value()) +
                                                 ".png" );
    //Crouched animation -> selected animation
    Texture *textureSelected = manager.getTexture( animationsRelPath +
                                                  std::string(crouchedNode.attribute("path").value()) +
                                                  ".png" );
    
    loadAnimation((*textureDefault), defaultNode, dstDefault);
    loadAnimation((*textureSelected), crouchedNode, dstSelected);
}
