/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Fight.hpp"
#include "Engines/GameEngine.hpp"

Fight::Fight(GameEngine *gameEngine) :
m_gameEngine(gameEngine), m_window(gameEngine->getWindow()),
m_configuration(gameEngine->getConfiguration()),
m_mode(FightMode::Default)
{
    
}

void Fight::setArena(const std::string &arenaName)
{
    m_arenaPath = arenaName;
}

void Fight::setMode(FightMode mode)
{
    m_mode = mode;
}

void Fight::addPlayer(const Character &character)
{
    m_playersDataList.push_back(character);
}

void Fight::clearPlayers()
{
    m_playersDataList.clear();
}

void Fight::run()
{
    //Init resource managers
    TextureManager textureManager;
    FontManager fontManager;
    SoundBufferManager soundBufferManager;
    
    float soundEngineVolume = static_cast<float>(m_configuration->getMasterVolume() * m_configuration->getSoundVolume()) / 100.f;
    
    SoundEngine soundEngine;
    soundEngine.loadSoundEffects("audio/sfx/sfx.xml");
    soundEngine.setManager(soundBufferManager);
    soundEngine.setVolume(soundEngineVolume);
    
    Vector viewSize = Vector(m_window->getViewWidth(), m_window->getViewHeight());
    
    //Interface
    EndTab endDialog(*fontManager.getFont("fonts/default.ttf", 30), textureManager, soundEngine);
    endDialog.setXPosition((viewSize.x - endDialog.getWidth()) / 2.f);
    endDialog.setYPosition((viewSize.y - endDialog.getHeight()) / 2.f);
    
    PauseTab pauseDialog(*fontManager.getFont("fonts/default.ttf", 30), textureManager, soundEngine);
    pauseDialog.setXPosition((viewSize.x - pauseDialog.getWidth()) / 2.f);
    pauseDialog.setYPosition((viewSize.y - pauseDialog.getHeight()) / 2.f);
    
    FightInterface interface;
    interface.viewWidth(viewSize.x);
    interface.setFonts(*fontManager.getFont("fonts/intuitive.ttf", 30),
                       *fontManager.getFont("fonts/default.ttf", 14),
                       *fontManager.getFont("fonts/default.ttf", 10));
    
    Text fpsText;
    fpsText.setFont(*fontManager.getFont("fonts/default.ttf", 18));
    fpsText.setPosition(viewSize.x - 70.f, 5.f);
    fpsText.setColor(Color(255, 255, 255));
    
    //Fight scene
    FightScene fightScene;
    fightScene.setMode(m_mode);
    fightScene.setArena(m_arenaPath);
    
    unsigned int humanID = 1;
    for (const Character &playerData : m_playersDataList)
    {
        if (playerData.type == Character::Type::Human)
        {
            int device = m_configuration->getPlayerDevice(humanID);
            PlayerKeys keys;
            
            if (device == DEVICE_KEYBOARD) //Keyboard
                m_configuration->getPlayerKeys(humanID, keys);
            else
                m_configuration->getPlayerJoystickKeys(humanID, m_gameEngine->getJoystickManager().getName(device), keys);
                    
            fightScene.addHumanPlayer(playerData.name, device, keys);
            humanID++;
        }
        else if (playerData.type == Character::Type::AI)
            fightScene.addAIPlayer(playerData.name);
    }
    
    fightScene.initialize(textureManager, soundEngine, m_configuration->getDebugMode());
    
    //Camera
    Camera interfaceCamera(Box(0.f, 0.f, viewSize.x, viewSize.y));
    FightCameraHandler fightCameraHdl = fightScene.getCameraHandler(m_configuration->isCameraDynamic() ?
                                                                    FightCameraHandler::CameraMode::Dynamic :
                                                                    FightCameraHandler::CameraMode::Static);
    //Player info interface
    interface.setPlayers(fightScene.getPlayers(), textureManager);
    
    //Start the music
    PlaylistPlayer::playlistNextSong(PLAYLIST_MAIN, true);
    
    //Start screen loop
    Timer time;
    Vector mouse;
    
    bool paused = false; //Is the game paused?
    bool finished = false; //Is the game finished?
    
    std::string winnerName;
    int winnerIx = -1;
    
    float frameTime = 0.f;
    
    float lastTime = Timer::getTicks(); //FPS counter
    float elapsedFrames = 0.f;
    
    SDL_Event event;
    while (1)
    {
        //Manage time
        frameTime = time.getTime();
        time.reset();
        
        //Manage events
        while (m_window->pollEvent(event))
        {
            switch (event.type)
            {
                case SDL_QUIT: //Application closed
                    m_gameEngine->exit();
                    return;
                break;
                
                case SDL_WINDOWEVENT:
                    if (event.window.event == SDL_WINDOWEVENT_RESIZED) //Window resized
                        m_window->resizeView();
                break;
                
                case SDL_JOYDEVICEADDED:
                    m_gameEngine->getJoystickManager().loadJoystick(event.jdevice.which);
                break;
                
                case SDL_JOYDEVICEREMOVED:
                    m_gameEngine->getJoystickManager().removeJoystick(event.jdevice.which);
                    m_configuration->updatePlayerDevices(event.jdevice.which);
                break;
                
                case SDL_KEYDOWN: //Key pressed
                    
                    //Pause handling
                    if (!finished && event.key.keysym.sym == SDLK_ESCAPE)
                    {
                        paused = !paused;
                        
                        //Mute / Unmute the sound engine (for unpleasant looping sounds especially)
                        soundEngine.setVolume( paused ? 0.f : soundEngineVolume );
                    }
                
                break;
                
                case SDL_MOUSEBUTTONUP: //Mouse released
                    mouse.x = static_cast<float>(event.button.x);
                    mouse.y = static_cast<float>(event.button.y);
                    
                    if (finished)
                        endDialog.mouseRelease(m_window->viewCoords(mouse));
                    else if (paused)
                        pauseDialog.mouseRelease(m_window->viewCoords(mouse));
                
                break;
                
                case SDL_MOUSEMOTION: //Mouse moved
                    mouse.x = static_cast<float>(event.motion.x);
                    mouse.y = static_cast<float>(event.motion.y);
                    
                    if (finished)
                        endDialog.mouseMove(m_window->viewCoords(mouse));
                    else if (paused)
                        pauseDialog.mouseMove(m_window->viewCoords(mouse));
                
                break;
                
                default:
                break;
            }
            
            //Process key events
            if (!paused) fightScene.processKeyEvent(event);
            fightScene.updateKeyState(event);
        }
        
        //End of fight
        if (!finished && fightScene.isFinished(winnerIx, winnerName))
        {
            if (winnerIx != -1)
                endDialog.setWinner(hnm::format("{} (#{}) wins!", winnerName, winnerIx+1));
            else
                endDialog.setWinner("Draw game!");
            
            finished = true;
        }
        
        if (finished && endDialog.chosen())
        {
            EndTab::Result result = endDialog.result();
            
            if (result == EndTab::Result::Exit)
            {
                PlaylistPlayer::playlistNextSong(PLAYLIST_MAIN, false);
                m_gameEngine->initMainMenu();
            }
            
            return;
        }
        
        if (paused && pauseDialog.chosen())
        {
            PauseTab::Result result = pauseDialog.result();
            
            if (result == PauseTab::Result::Resume)
            {
                paused = false;
                soundEngine.setVolume(soundEngineVolume);
                
                pauseDialog.reinitialize();
            }
            else if (result == PauseTab::Result::Restart)
                return; //Game engine will restart current screen.
            else if (result == PauseTab::Result::Exit)
            {
                PlaylistPlayer::playlistNextSong(PLAYLIST_MAIN, false);
                m_gameEngine->initMainMenu();
                return;
            }
        }
        
        //Update fight scene
        if (!paused) fightScene.update(frameTime);
        
        //Update sounds & interface
        soundEngine.update();
        interface.update();
        
        //FPS counter
        elapsedFrames += 1.f;
        if (Timer::getTicks() - lastTime >= 250.f)
        {
            int framerate = static_cast<int>( (elapsedFrames / (Timer::getTicks() - lastTime)) * 1000 );
            fpsText.setText(hnm::format("FPS: {}", framerate));
            
            lastTime = Timer::getTicks();
            elapsedFrames = 0.f;
        }
        
        //Display
            //Clear
        m_window->clear();
        
            //Graphic part
        fightCameraHdl.look(viewSize.x, viewSize.y, frameTime, fightScene.getMinCameraWidth());
        fightScene.draw();
        
            //Interface part
        interfaceCamera.look(viewSize.x, viewSize.y);
        
        if (finished) endDialog.draw();
        else if (paused) pauseDialog.draw();
        
        interface.draw();
        fpsText.draw();
        
            //Render
        m_window->flush();
    }
}
