/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "FightInterface.hpp"

FightInterface::FightInterface() :
Drawable(), m_namesFont(nullptr), m_weaponsFont(nullptr), m_barsFont(nullptr), m_viewWidth(-1.f), m_iconsScale(1.f)
{
    
}

void FightInterface::draw()
{
    //UpdateMatrix
    Drawable::pushMatrix();
    Drawable::updateMatrix();
    
    for (PlayerInfo &playerInfo : m_playerInfos)
        playerInfo.draw();
    
    Drawable::popMatrix();
}

void FightInterface::update()
{
    for (PlayerInfo &playerInfo : m_playerInfos)
        playerInfo.update();
}

void FightInterface::viewWidth(float width)
{
    m_viewWidth = width;
}

void FightInterface::setPlayers(const std::vector<const Player*> &players, TextureManager &textureManager)
{
    m_playerInfos.clear();
    
    for (const Player *player : players)
    {
        if (player != nullptr)
        {
            PlayerInfo playerInfo;
            playerInfo.setFonts(*m_namesFont, *m_weaponsFont, *m_barsFont);
            playerInfo.setPlayer((*player), textureManager);
            playerInfo.setIconScale(m_iconsScale);
            
            m_playerInfos.push_back(playerInfo);
        }
    }
    
    updatePositions();
}

void FightInterface::updatePositions()
{
    //Compute the space between each PlayerInfo.
    float space = 50.f;
    if (m_viewWidth > 0.f)
    {
        //Compute content width
        float contentWidth = 0.f;
        for (PlayerInfo &playerInfo : m_playerInfos)
            contentWidth += playerInfo.getWidth();
        
        //Reduce size of items if they overflow
        for (unsigned int resizeAttempts = 0; resizeAttempts < 2; resizeAttempts++)
        {
            //Break if size is sufficient
            if (contentWidth <= m_viewWidth)
                break;
            
            //Resize the player info items
            m_iconsScale *= .75f;
            for (PlayerInfo &playerInfo : m_playerInfos)
                playerInfo.setIconScale(m_iconsScale);
            
            //Recompute contentWidth
            contentWidth = 0.f;
            for (PlayerInfo &playerInfo : m_playerInfos)
                contentWidth += playerInfo.getWidth();
        }
        
        space = (m_viewWidth - contentWidth) / static_cast<float>(m_playerInfos.size() + 1);
    }
    
    //Position each PlayerInfo
    float currentPos = space;
    for (PlayerInfo &playerInfo : m_playerInfos)
    {
        playerInfo.setXPosition(currentPos);
        currentPos += playerInfo.getWidth() + space;
    }
}

//Get and set methods
void FightInterface::setFonts(Font &namesFont, Font &weaponsFont, Font &barsFont)
{
    m_namesFont = &namesFont;
    m_weaponsFont = &weaponsFont;
    m_barsFont = &barsFont;
    
    for (PlayerInfo &playerInfo : m_playerInfos)
        playerInfo.setFonts(*m_namesFont, *m_weaponsFont, *m_barsFont);
}
