/*Copyright (C) 2010-2022 Duncan Deveaux

This file is part of Hikou no mizu.

Hikou no mizu is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Hikou no mizu is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Hikou no mizu.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "FightScene.hpp"

FightScene::FightScene() : m_showHitboxes(false), m_weapons(nullptr)
{
    
}

FightScene::~FightScene()
{
    if (m_weapons != nullptr)
        delete m_weapons;
}

void FightScene::setMode(FightMode mode)
{
    m_mode = mode;
}

void FightScene::setArena(const std::string &arenaPath)
{
    m_arenaPath = arenaPath;
}

void FightScene::addAIPlayer(const std::string &name)
{
    m_aisList.push_back(AI::loadFromXML(name, "cfg/characters.xml"));
    m_players.push_back(&m_aisList.back());
}

void FightScene::addHumanPlayer(const std::string &name, int deviceID, const PlayerKeys &keys)
{
    Human human = Human::loadFromXML(name, "cfg/characters.xml");
    human.setDeviceID(deviceID);
    human.setKeys(keys);
    
    m_humansList.push_back(human);
    m_players.push_back(&m_humansList.back());
}

void FightScene::initialize(TextureManager &textureManager, SoundEngine &soundEngine, bool showHitboxes)
{
    ///Physics world and arena
    m_showHitboxes = showHitboxes;
    m_physicsWorld = PhysicsWorld();
    
    m_arena.loadFromXML(m_arenaPath, "cfg/arenas.xml", textureManager);
    m_arena.applyToPhysics(m_physicsWorld);
    m_arena.subscribeObjects(m_physicsWorld, &soundEngine);
    
    ///Players
    for (unsigned int i = 0; i < m_players.size(); i++)
    {
        Player *player = m_players[i];
        
        //General initialization
        player->setPhysicsWorld(m_physicsWorld);
        player->setSoundEngine(soundEngine);
        player->setDebugHits(m_showHitboxes);
        
        player->loadGraphics(textureManager);
        
        //Spawn
        Vector playerSpawn(static_cast<float>(i) * 300.f, player->getBox().height);
        m_arena.getPlayerSpawn(i, playerSpawn);
        
        player->setPosition(playerSpawn.x, playerSpawn.y - player->getBox().height);
        if (playerSpawn.x > m_physicsWorld.getWorldBox().width / 2.f)
            player->setXFlip(true);
        
        //Subscribe to physics
        m_physicsWorld.addObject((*player));
    }
    
    //Informing AIs about who their enemies are
    for (AI &ai : m_aisList)
    {
        //Everybody else is an enemy ;)
        for (Player *player : m_players)
        {
            if (player != &ai)
                ai.addEnemy(*player);
        }
    }
    
    ///Weapons
    if (m_weapons != nullptr)
        delete m_weapons;
    
    std::vector<const Player*> const_players = getPlayers();
    m_weapons = new WeaponMedium(const_players, textureManager, soundEngine);
    
    for (Player *player : m_players)
        player->initWeapons((*m_weapons));
    
    ///Versus mode
    if (m_mode == FightMode::Versus && m_players.size() == 2)
    {
        m_players[0]->lockEnnemy(const_players[1]);
        m_players[1]->lockEnnemy(const_players[0]);
    }
}

void FightScene::processKeyEvent(const SDL_Event &event)
{
    //Trigger event on pressed/unpressed keys
    for (Human &human : m_humansList)
    {
        if (!human.isKo() && !human.isHit())
            human.keyEvent(event);
    }
}

void FightScene::updateKeyState(const SDL_Event &event)
{
    //Update key state based on
    //the pressed/unpressed keyboard or joystick keys
    for (Human &human : m_humansList)
        human.updateKeyState(event);
}

void FightScene::update(float frameTime)
{
    m_physicsWorld.update(frameTime);
    
    if (m_weapons != nullptr)
        m_weapons->update(frameTime);
    
    for (Player *player : m_players)
    {
        player->act();
        player->update(frameTime);
    }
}

void FightScene::draw()
{
    m_arena.draw();
    
    for (Player *player : m_players)
        player->draw();
    
    if (m_weapons != nullptr)
        m_weapons->drawWeapons();
    
    if (m_showHitboxes)
        m_physicsWorld.debug();
}

bool FightScene::isFinished(int &winnerIxDst, std::string &winnerNameDst) const
{
    int notKO = -1; //To be set to the index of a not KO player
    for (unsigned int i = 0; i < m_players.size(); i++)
    {
        const Player *player = m_players[i];
        if (!player->isKo())
        {
            if (notKO != -1) return false; //At least two players are still not KO
            else notKO = static_cast<int>(i);
        }
    }
    
    winnerIxDst = notKO;
    if (winnerIxDst != -1) //Not a draw game. (Draw games should not be possible at the moment)
        winnerNameDst = m_players[static_cast<unsigned int>(winnerIxDst)]->getName();
    
    return true;
}

std::vector<const Player*> FightScene::getPlayers() const
{
    std::vector<const Player*> res;
    
    for (const Player *player : m_players)
        res.push_back(player);
    
    return res;
}

FightCameraHandler FightScene::getCameraHandler(FightCameraHandler::CameraMode cameraMode) const
{
    return FightCameraHandler(Box(0.f, 0.f, m_arena.getWidth(), m_arena.getHeight()),
                              m_arena.getGround(),
                              getPlayers(),
                              cameraMode);
}

float FightScene::getMinCameraWidth() const
{
    return m_arena.getMinCameraWidth();
}
